# #
# Copyright 2015-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #

"""
Support for checking types of easyconfig parameter values.

:author: Caroline De Brouwer (Ghent University)
:author: Kenneth Hoste (Ghent University)
"""
from distutils.util import strtobool

from easybuild.base import fancylogger
from easybuild.framework.easyconfig.format.format import DEPENDENCY_PARAMETERS
from easybuild.framework.easyconfig.format.format import SANITY_CHECK_PATHS_DIRS, SANITY_CHECK_PATHS_FILES
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.py2vs3 import string_type

_log = fancylogger.getLogger('easyconfig.types', fname=False)


def as_hashable(dict_value):
    """Helper function, convert dict value to hashable equivalent via tuples."""
    res = []
    for key, val in sorted(dict_value.items()):
        if isinstance(val, list):
            val = tuple(val)
        elif isinstance(val, dict):
            val = as_hashable(val)
        res.append((key, val))
    return tuple(res)


def check_element_types(elems, allowed_types):
    """
    Check whether types of elements of specified (iterable) value are as expected.

    :param elems: iterable value (list or dict) of elements
    :param allowed_types: allowed types per element; either a simple list, or a dict of allowed_types by element name
    """
    # combine elements with their list of allowed types
    elems_and_allowed_types = None
    if isinstance(elems, (list, tuple)):
        if isinstance(allowed_types, (list, tuple)):
            elems_and_allowed_types = [(elem, allowed_types) for elem in elems]
        else:
            raise EasyBuildError("Don't know how to combine value of type %s with allowed types of type %s",
                                 type(elems), type(allowed_types))
    elif isinstance(elems, dict):
        # allowed_types can be a tuple representation of a dict, or a flat list of types

        # try to convert to a dict, but ignore if it fails
        try:
            allowed_types = dict(allowed_types)
        except (ValueError, TypeError):
            pass

        if isinstance(allowed_types, (list, tuple)):
            elems_and_allowed_types = [(elem, allowed_types) for elem in elems.values()]
        elif isinstance(allowed_types, dict):
            elems_and_allowed_types = []
            for key, val in elems.items():
                if key in allowed_types:
                    elems_and_allowed_types.append((val, allowed_types[key]))
                else:
                    # if key has no known allowed types, use empty list of allowed types to yield False check result
                    elems_and_allowed_types.append((val, []))
        else:
            raise EasyBuildError("Unknown type of allowed types specification: %s", type(allowed_types))
    else:
        raise EasyBuildError("Don't know how to check element types for value of type %s: %s", type(elems), elems)

    # check whether all element types are allowed types
    res = True
    for elem, allowed_types_elem in elems_and_allowed_types:
        res &= any(is_value_of_type(elem, t) for t in allowed_types_elem)

    return res


def check_key_types(val, allowed_types):
    """Check whether type of keys for specific dict value are as expected."""
    if isinstance(val, dict):
        res = True
        for key in val.keys():
            res &= any(is_value_of_type(key, t) for t in allowed_types)
    else:
        _log.debug("Specified value %s (type: %s) is not a dict, so key types check failed", val, type(val))
        res = False

    return res


def check_known_keys(val, allowed_keys):
    """Check whether all keys for specified dict value are known keys."""
    if isinstance(val, dict):
        res = all(key in allowed_keys for key in val.keys())
    else:
        _log.debug("Specified value %s (type: %s) is not a dict, so known keys check failed", val, type(val))
        res = False
    return res


def check_required_keys(val, required_keys):
    """Check whether all required keys are present in the specified dict value."""
    if isinstance(val, dict):
        keys = val.keys()
        res = all(key in keys for key in required_keys)
    else:
        _log.debug("Specified value %s (type: %s) is not a dict, so known keys check failed", val, type(val))
        res = False
    return res


def is_value_of_type(value, expected_type):
    """
    Check whether specified value matches a particular very specific (non-trivial) type,
    which is specified by means of a 2-tuple: (parent type, tuple with additional type requirements).

    :param value: value to check the type of
    :param expected_type: type of value to check against
    """
    type_ok = False

    if expected_type in EASY_TYPES:
        # easy types can be checked using isinstance
        type_ok = isinstance(value, expected_type)

    elif expected_type in CHECKABLE_TYPES:
        # more complex types need to be checked differently, through helper functions for extra type requirements
        parent_type = expected_type[0]
        extra_reqs = dict(expected_type[1])

        # first step: check parent type
        type_ok = isinstance(value, parent_type)
        if type_ok:
            _log.debug("Parent type of value %s matches %s, going in...", value, parent_type)
            # second step: check additional type requirements
            extra_req_checkers = {
                'elem_types': lambda val: check_element_types(val, extra_reqs['elem_types']),
            }
            if parent_type == dict:
                extra_req_checkers.update({
                    'key_types': lambda val: check_key_types(val, extra_reqs['key_types']),
                    'opt_keys': lambda val: check_known_keys(val, extra_reqs['opt_keys'] + extra_reqs['req_keys']),
                    'req_keys': lambda val: check_required_keys(val, extra_reqs['req_keys']),
                })

            for er_key in extra_reqs:
                if er_key in extra_req_checkers:
                    check_ok = extra_req_checkers[er_key](value)
                    msg = ('FAILED', 'passed')[check_ok]
                    type_ok &= check_ok
                    _log.debug("Check for %s requirement (%s) %s for %s", er_key, extra_reqs[er_key], msg, value)
                else:
                    raise EasyBuildError("Unknown type requirement specified: %s", er_key)

            msg = ('FAILED', 'passed')[type_ok]
            _log.debug("Non-trivial value type checking of easyconfig value '%s': %s", value, msg)

        else:
            _log.debug("Parent type of value %s doesn't match %s: %s", value, parent_type, type(value))

    else:
        raise EasyBuildError("Don't know how to check whether specified value is of type %s", expected_type)

    return type_ok


def check_type_of_param_value(key, val, auto_convert=False):
    """
    Check value type of specified easyconfig parameter.

    :param key: name of easyconfig parameter
    :param val: easyconfig parameter value, of which type should be checked
    :param auto_convert: try to automatically convert to expected value type if required
    """
    type_ok, newval = False, None
    expected_type = PARAMETER_TYPES.get(key)

    # check value type
    if expected_type is None:
        _log.debug("No type specified for easyconfig parameter '%s', so skipping type check.", key)
        type_ok = True

    else:
        type_ok = is_value_of_type(val, expected_type)

    # determine return value, attempt type conversion if needed/requested
    if type_ok:
        _log.debug("Value type check passed for %s parameter value: %s", key, val)
        newval = val
    elif auto_convert:
        _log.debug("Value type check for %s parameter value failed, going to try to automatically convert to %s",
                   key, expected_type)
        # convert_value_type will raise an error if the conversion fails
        newval = convert_value_type(val, expected_type)
        type_ok = True
    else:
        _log.debug("Value type check for %s parameter value failed, auto-conversion of type not enabled", key)

    return type_ok, newval


def convert_value_type(val, typ):
    """
    Try to convert type of provided value to specific type.

    :param val: value to convert type of
    :param typ: target type
    """
    res = None

    if typ in EASY_TYPES and isinstance(val, typ):
        _log.debug("Value %s is already of specified target type %s, no conversion needed", val, typ)
        res = val

    elif typ in CHECKABLE_TYPES and is_value_of_type(val, typ):
        _log.debug("Value %s is already of specified non-trivial target type %s, no conversion needed", val, typ)
        res = val

    elif typ in TYPE_CONVERSION_FUNCTIONS:
        func = TYPE_CONVERSION_FUNCTIONS[typ]
        _log.debug("Trying to convert value %s (type: %s) to %s using %s", val, type(val), typ, func)
        try:
            res = func(val)
            _log.debug("Type conversion seems to have worked, new type: %s", type(res))
        except Exception as err:
            raise EasyBuildError("Converting type of %s (%s) to %s using %s failed: %s", val, type(val), typ, func, err)

        if not isinstance(res, typ):
            raise EasyBuildError("Converting value %s to type %s didn't work as expected: got %s", val, typ, type(res))

    else:
        raise EasyBuildError("No conversion function available (yet) for target type %s", typ)

    return res


def to_toolchain_dict(spec):
    """
    Convert a comma-separated string or 2/3-element list of strings to a dictionary with name/version keys, and
    optionally a hidden key. If the specified value is a dict already, the keys are checked to be only
    name/version/hidden.

    For example: "intel, 2015a" => {'name': 'intel', 'version': '2015a'}
                 "foss, 2016a, True" => {'name': 'foss', 'version': '2016a', 'hidden': True}

    :param spec: a comma-separated string with two or three values, or a 2/3-element list of strings, or a dict
    """
    # check if spec is a string or a list of two values; else, it can not be converted
    if isinstance(spec, string_type):
        spec = spec.split(',')

    if isinstance(spec, (list, tuple)):
        # 2-element list
        if len(spec) == 2:
            res = {'name': spec[0].strip(), 'version': spec[1].strip()}
        # 3-element list
        elif len(spec) == 3:
            res = {'name': spec[0].strip(), 'version': spec[1].strip(), 'hidden': strtobool(spec[2].strip())}
        else:
            raise EasyBuildError("Can not convert list %s to toolchain dict. Expected 2 or 3 elements", spec)

    elif isinstance(spec, dict):
        # already a dict, check keys
        sorted_keys = sorted(spec.keys())
        if sorted_keys == ['name', 'version'] or sorted_keys == ['hidden', 'name', 'version']:
            res = spec
        else:
            raise EasyBuildError("Incorrect set of keys in provided dictionary, should be only name/version/hidden: %s",
                                 spec)

    else:
        raise EasyBuildError("Conversion of %s (type %s) to toolchain dict is not supported", spec, type(spec))

    return res


def to_name_version_dict(spec):
    """No longer supported, replaced by to_toolchain_dict."""
    _log.nosupport("to_name_version_dict; use to_toolchain_dict instead.", '3.0')


def to_list_of_strings(value):
    """
    Convert specified value to a list of strings, if possible.

    Supported: single string value, tuple of string values.
    """
    res = None

    # if value is already of correct type, we don't need to change anything
    if isinstance(value, list) and all(isinstance(s, string_type) for s in value):
        res = value
    elif isinstance(value, string_type):
        res = [value]
    elif isinstance(value, tuple) and all(isinstance(s, string_type) for s in value):
        res = list(value)
    else:
        raise EasyBuildError("Don't know how to convert provided value to a list of strings: %s", value)

    return res


def to_list_of_strings_and_tuples(spec):
    """
    Convert a 'list of lists and strings' to a 'list of tuples and strings'

    Example:
        ['foo', ['bar', 'baz']]
        to
        ['foo', ('bar', 'baz')]
    """
    str_tup_list = []

    if not isinstance(spec, (list, tuple)):
        raise EasyBuildError("Expected value to be a list, found %s (%s)", spec, type(spec))

    for elem in spec:
        if isinstance(elem, (string_type, tuple)):
            str_tup_list.append(elem)
        elif isinstance(elem, list):
            str_tup_list.append(tuple(elem))
        else:
            raise EasyBuildError("Expected elements to be of type string, tuple or list, got %s (%s)", elem, type(elem))

    return str_tup_list


def to_sanity_check_paths_dict(spec):
    """
    Convert a sanity_check_paths dict as received by yaml (a dict with list values that contain either lists or strings)

    Example:
        {'files': ['file1', ['file2a', 'file2b]], 'dirs': ['foo/bar']}
        to
        {'files': ['file1', ('file2a', 'file2b')], 'dirs': ['foo/bar']}
    """
    if not isinstance(spec, dict):
        raise EasyBuildError("Expected value to be a dict, found %s (%s)", spec, type(spec))

    sanity_check_dict = {}
    for key in spec:
        sanity_check_dict[key] = to_list_of_strings_and_tuples(spec[key])
    return sanity_check_dict


# this uses to_toolchain, so it needs to be at the bottom of the module
def to_dependency(dep):
    """
    Convert a dependency specification to a dependency dict with name/version/versionsuffix/toolchain keys.

    Example:
        {'foo': '1.2.3', 'toolchain': 'GCC, 4.8.2'}
        to
        {'name': 'foo', 'version': '1.2.3', 'toolchain': {'name': 'GCC', 'version': '4.8.2'}}

    or
        {'name': 'fftw/3.3.4.1', 'external_module': True}
        to
        {'name': 'fftw/3.3.4.1', 'external_module': True, 'version': None}
    """
    # deal with dependencies coming for .eb easyconfig, typically in tuple format:
    #   (name, version[, versionsuffix[, toolchain]])
    if isinstance(dep, dict):
        depspec = {}

        if dep.get('external_module', False):
            expected_keys = ['external_module', 'name']
            if sorted(dep.keys()) == expected_keys:
                depspec.update({
                    'external_module': True,
                    'full_mod_name': dep['name'],
                    'name': None,
                    'short_mod_name': dep['name'],
                    'version': None,
                })
            else:
                raise EasyBuildError("Unexpected format for dependency marked as external module: %s", dep)

        else:
            dep_keys = list(dep.keys())

            # need to handle name/version keys first, to avoid relying on order in which keys are processed...
            for key in ['name', 'version']:
                if key in dep:
                    depspec[key] = str(dep[key])
                    dep_keys.remove(key)

            for key in dep_keys:
                if key == 'versionsuffix':
                    depspec[key] = str(dep[key])
                elif key == 'toolchain':
                    depspec['toolchain'] = to_toolchain_dict(dep[key])
                elif not ('name' in depspec and 'version' in depspec):
                    depspec.update({'name': key, 'version': str(dep[key])})
                else:
                    raise EasyBuildError("Found unexpected (key, value) pair: %s, %s", key, dep[key])

            if not ('name' in depspec and 'version' in depspec):
                raise EasyBuildError("Can not parse dependency without name and version: %s", dep)

    else:
        # pass down value untouched, let EasyConfig._parse_dependency handle it
        depspec = dep
        if isinstance(dep, (tuple, list)):
            _log.debug("Passing down dependency value of type %s without touching it: %s", type(dep), dep)
        else:
            _log.warning("Unknown type of value in to_dependency %s; passing value down as is: %s", type(dep), dep)

    return depspec


def to_dependencies(dep_list):
    """
    Convert a list of dependencies obtained from parsing a .yeb easyconfig
    to a list of dependencies in the correct format
    """
    return [to_dependency(dep) for dep in dep_list]


def to_checksums(checksums):
    """Ensure correct element types for list of checksums: convert list elements to tuples."""
    res = []
    for checksum in checksums:
        # each list entry can be:
        # * None (indicates no checksum)
        # * a string (MD5 or SHA256 checksum)
        # * a tuple with 2 elements: checksum type + checksum value
        # * a list of checksums (i.e. multiple checksums for a single file)
        # * a dict (filename to checksum mapping)
        if isinstance(checksum, string_type):
            res.append(checksum)
        elif isinstance(checksum, (list, tuple)):
            # 2 elements + only string/int values => a checksum tuple
            if len(checksum) == 2 and all(isinstance(x, (string_type, int)) for x in checksum):
                res.append(tuple(checksum))
            else:
                res.append(to_checksums(checksum))
        elif isinstance(checksum, dict):
            validated_dict = {}
            for key, value in checksum.items():
                validated_dict[key] = to_checksums(value)
            res.append(validated_dict)
        else:
            res.append(checksum)

    return res


def ensure_iterable_license_specs(specs):
    """
    Ensures that the provided license file/server specifications are of correct type and converts
    them to a list.  The input can either be None, a string, or a list/tuple of strings.

    :param specs: License file/server specifications as provided via `license_file` easyconfig parameter
    """
    if specs is None:
        license_specs = [None]
    elif isinstance(specs, string_type):
        license_specs = [specs]
    elif isinstance(specs, (list, tuple)) and all(isinstance(x, string_type) for x in specs):
        license_specs = list(specs)
    else:
        msg = "Unsupported type %s for easyconfig parameter 'license_file'! " % type(specs)
        msg += "Can either be None, a string, or a tuple/list of strings."
        raise EasyBuildError(msg)

    return license_specs


# these constants use functions defined in this module, so they needs to be at the bottom of the module
# specific type: dict with only name/version as keys with string values, and optionally a hidden key with bool value
# additional type requirements are specified as tuple of tuples rather than a dict, since this needs to be hashable
TOOLCHAIN_DICT = (dict, as_hashable({
    'elem_types': {
        'hidden': [bool],
        'name': [str],
        'version': [str],
    },
    'opt_keys': ['hidden'],
    'req_keys': ['name', 'version'],
}))
DEPENDENCY_DICT = (dict, as_hashable({
    'elem_types': {
        'full_mod_name': [str],
        'name': [str],
        'short_mod_name': [str],
        'toolchain': [TOOLCHAIN_DICT],
        'version': [str],
        'versionsuffix': [str],
    },
    'opt_keys': ['full_mod_name', 'short_mod_name', 'toolchain', 'versionsuffix'],
    'req_keys': ['name', 'version'],
}))
DEPENDENCIES = (list, as_hashable({'elem_types': [DEPENDENCY_DICT]}))

TUPLE_OF_STRINGS = (tuple, as_hashable({'elem_types': [str]}))
LIST_OF_STRINGS = (list, as_hashable({'elem_types': [str]}))
STRING_OR_TUPLE_LIST = (list, as_hashable({'elem_types': [str, TUPLE_OF_STRINGS]}))
STRING_DICT = (dict, as_hashable(
    {
        'elem_types': [str],
        'key_types': [str],
    }
))
SANITY_CHECK_PATHS_DICT = (dict, as_hashable({
    'elem_types': {
        SANITY_CHECK_PATHS_FILES: [STRING_OR_TUPLE_LIST],
        SANITY_CHECK_PATHS_DIRS: [STRING_OR_TUPLE_LIST],
    },
    'opt_keys': [],
    'req_keys': [SANITY_CHECK_PATHS_FILES, SANITY_CHECK_PATHS_DIRS],
}))
# checksums is a list of checksums, one entry per file (source/patch)
# each entry can be:
# a single checksum value (string)
# a single checksum value of a specified type (2-tuple, 1st element is checksum type, 2nd element is checksum)
# a list of checksums (of different types, perhaps different formats), which should *all* be valid
# a dictionary with a mapping from filename to checksum value
CHECKSUM_LIST = (list, as_hashable({'elem_types': [str, tuple, STRING_DICT]}))
CHECKSUMS = (list, as_hashable({'elem_types': [str, tuple, STRING_DICT, CHECKSUM_LIST]}))

CHECKABLE_TYPES = [CHECKSUM_LIST, CHECKSUMS, DEPENDENCIES, DEPENDENCY_DICT, LIST_OF_STRINGS,
                   SANITY_CHECK_PATHS_DICT, STRING_DICT, STRING_OR_TUPLE_LIST, TOOLCHAIN_DICT, TUPLE_OF_STRINGS]

# easy types, that can be verified with isinstance
EASY_TYPES = [string_type, bool, dict, int, list, str, tuple]

# type checking is skipped for easyconfig parameters names not listed in PARAMETER_TYPES
PARAMETER_TYPES = {
    'checksums': CHECKSUMS,
    'docurls': LIST_OF_STRINGS,
    'name': string_type,
    'osdependencies': STRING_OR_TUPLE_LIST,
    'patches': STRING_OR_TUPLE_LIST,
    'sanity_check_paths': SANITY_CHECK_PATHS_DICT,
    'toolchain': TOOLCHAIN_DICT,
    'version': string_type,
}
# add all dependency types as dependencies
for dep in DEPENDENCY_PARAMETERS:
    PARAMETER_TYPES[dep] = DEPENDENCIES

TYPE_CONVERSION_FUNCTIONS = {
    string_type: str,
    float: float,
    int: int,
    str: str,
    CHECKSUMS: to_checksums,
    DEPENDENCIES: to_dependencies,
    LIST_OF_STRINGS: to_list_of_strings,
    TOOLCHAIN_DICT: to_toolchain_dict,
    SANITY_CHECK_PATHS_DICT: to_sanity_check_paths_dict,
    STRING_OR_TUPLE_LIST: to_list_of_strings_and_tuples,
}



# #
# Copyright 2015-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Unit tests for packaging support.

@author: Kenneth Hoste (Ghent University)
"""
import os
import re
import stat
import sys

from test.framework.utilities import EnhancedTestCase, TestLoaderFiltered, init_config
from unittest import TextTestRunner

from easybuild.framework.easyconfig.easyconfig import EasyConfig
from easybuild.tools.config import log_path
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, read_file, write_file
from easybuild.tools.package.utilities import ActivePNS, avail_package_naming_schemes, check_pkg_support, package
from easybuild.tools.version import VERSION as EASYBUILD_VERSION

FPM_OUTPUT_FILE = 'fpm_mocked.out'

# purposely using non-bash script, to detect issues with shebang line being ignored (run_cmd with shell=False)
MOCKED_FPM = """#!/usr/bin/env python
import os, sys

def verbose(msg):
    with open('%(fpm_output_file)s', 'a') as fp:
        fp.write(msg + '\\n')

description, iteration, name, source, target, url, version, workdir = '', '', '', '', '', '', '', ''
excludes = []

verbose(' '.join(sys.argv[1:]))

idx = 1
while idx < len(sys.argv):

    if sys.argv[idx] == '--workdir':
        idx += 1
        workdir = sys.argv[idx]
        verbose('workdir'); verbose(workdir)

    elif sys.argv[idx] == '--name':
        idx += 1
        name = sys.argv[idx]

    elif sys.argv[idx] == '--version':
        idx += 1
        version = sys.argv[idx]
        verbose('version'); verbose(version)

    elif sys.argv[idx] == '--description':
        idx += 1
        description = sys.argv[idx]

    elif sys.argv[idx] == '--url':
        idx += 1
        url = sys.argv[idx]

    elif sys.argv[idx] == '--iteration':
        idx += 1
        iteration = sys.argv[idx]

    elif sys.argv[idx] == '-t':
        idx += 1
        target = sys.argv[idx]

    elif sys.argv[idx] == '-s':
        idx += 1
        source = sys.argv[idx]

    elif sys.argv[idx] == '--exclude':
        idx += 1
        excludes.append(sys.argv[idx])

    elif sys.argv[idx].startswith('--'):
        verbose("got an unhandled option: " + sys.argv[idx] + ' ' + sys.argv[idx+1])
        idx += 1

    else:
        installdir = sys.argv[idx]
        modulefile = sys.argv[idx+1]
        break

    idx += 1

pkgfile = os.path.join(workdir, name + '-' + version + '.' + iteration + '.' + target)

with open(pkgfile, 'w') as fp:
    fp.write('thisisan' + target + '\\n')
    fp.write(' '.join(sys.argv[1:]) + '\\n')
    fp.write("STARTCONTENTS of installdir " + installdir + ':\\n')

    find_cmd = 'find ' + installdir + '  ' + ''.join([" -not -path /" + x + ' ' for x in excludes])
    verbose("trying: " + find_cmd)
    fp.write(find_cmd + '\\n')

    fp.write('ENDCONTENTS\\n')

    fp.write("Contents of module file " + modulefile + ':')


    fp.write('modulefile: ' + modulefile + '\\n')
    #modtxt = open(modulefile).read()
    #fp.write(modtxt + '\\n')

    fp.write("I found excludes " + ' '.join(excludes) + '\\n')
    fp.write("DESCRIPTION: " + description + '\\n')
"""


def mock_fpm(tmpdir):
    """Put mocked version of fpm command in place in specified tmpdir."""
    # put mocked 'fpm' command in place, just for testing purposes
    fpm = os.path.join(tmpdir, 'fpm')
    write_file(fpm, MOCKED_FPM % {'fpm_output_file': os.path.join(tmpdir, FPM_OUTPUT_FILE)})
    adjust_permissions(fpm, stat.S_IXUSR, add=True)

    # also put mocked rpmbuild in place
    rpmbuild = os.path.join(tmpdir, 'rpmbuild')
    write_file(rpmbuild, '#!/bin/bash')  # only needs to be there, doesn't need to actually do something...
    adjust_permissions(rpmbuild, stat.S_IXUSR, add=True)

    os.environ['PATH'] = '%s:%s' % (tmpdir, os.environ['PATH'])


class PackageTest(EnhancedTestCase):
    """Tests for packaging support."""

    def test_avail_package_naming_schemes(self):
        """Test avail_package_naming_schemes()"""
        self.assertEqual(sorted(avail_package_naming_schemes().keys()), ['EasyBuildPNS'])

    def test_check_pkg_support(self):
        """Test check_pkg_support()."""

        # clear $PATH to make sure fpm/rpmbuild can not be found
        os.environ['PATH'] = ''

        self.assertErrorRegex(EasyBuildError, "Selected packaging tool 'fpm' not found", check_pkg_support)

        for binary in ['fpm', 'rpmbuild']:
            binpath = os.path.join(self.test_prefix, binary)
            write_file(binpath, '#!/bin/bash')
            adjust_permissions(binpath, stat.S_IXUSR, add=True)
        os.environ['PATH'] = self.test_prefix

        # no errors => support check passes
        check_pkg_support()

    def test_active_pns(self):
        """Test use of ActivePNS."""
        init_config(build_options={'silent': True})

        topdir = os.path.dirname(os.path.abspath(__file__))
        test_easyconfigs = os.path.join(topdir, 'easyconfigs', 'test_ecs')
        test_ec = os.path.join(test_easyconfigs, 'o', 'OpenMPI', 'OpenMPI-2.1.2-GCC-6.4.0-2.28.eb')
        ec = EasyConfig(test_ec, validate=False)

        pns = ActivePNS()

        # default: EasyBuild package naming scheme, pkg release 1
        self.assertEqual(pns.name(ec), 'OpenMPI-2.1.2-GCC-6.4.0-2.28')
        self.assertEqual(pns.version(ec), 'eb-%s' % EASYBUILD_VERSION)
        self.assertEqual(pns.release(ec), '1')

    def test_package(self):
        """Test package function."""
        build_options = {
            'package_tool_options': '--foo bar',
            'silent': True,
        }
        init_config(build_options=build_options)

        topdir = os.path.dirname(os.path.abspath(__file__))
        test_easyconfigs = os.path.join(topdir, 'easyconfigs', 'test_ecs')
        ec = EasyConfig(os.path.join(test_easyconfigs, 't', 'toy', 'toy-0.0-gompi-2018a-test.eb'), validate=False)

        mock_fpm(self.test_prefix)

        # import needs to be done here, since test easyblocks are only included later
        from easybuild.easyblocks.toy import EB_toy
        easyblock = EB_toy(ec)

        # build & install first
        easyblock.run_all_steps(False)

        # write a dummy log and report file to make sure they don't get packaged
        logfile = os.path.join(easyblock.installdir, log_path(), "logfile.log")
        write_file(logfile, "I'm a logfile")
        reportfile = os.path.join(easyblock.installdir, log_path(), "report.md")
        write_file(reportfile, "I'm a reportfile")

        # package using default packaging configuration (FPM to build RPM packages)
        pkgdir = package(easyblock)
        pkgfile = os.path.join(pkgdir, 'toy-0.0-gompi-2018a-test-eb-%s.1.rpm' % EASYBUILD_VERSION)

        fpm_output = read_file(os.path.join(self.test_prefix, FPM_OUTPUT_FILE))
        pkgtxt = read_file(pkgfile)

        self.assertTrue(os.path.isfile(pkgfile), "Found %s" % pkgfile)

        # check whether extra packaging options were passed down
        regex = re.compile("^got an unhandled option: --foo bar$", re.M)
        self.assertTrue(regex.search(fpm_output), "Pattern '%s' found in: %s" % (regex.pattern, fpm_output))

        pkgtxt = read_file(pkgfile)
        pkgtxt_regex = re.compile("STARTCONTENTS of installdir %s" % easyblock.installdir)
        self.assertTrue(pkgtxt_regex.search(pkgtxt), "Pattern '%s' found in: %s" % (pkgtxt_regex.pattern, pkgtxt))

        no_logfiles_regex = re.compile(r'STARTCONTENTS.*\.(log|md)$.*ENDCONTENTS', re.DOTALL | re.MULTILINE)
        res = no_logfiles_regex.search(pkgtxt)
        self.assertFalse(res, "Pattern not '%s' found in: %s" % (no_logfiles_regex.pattern, pkgtxt))

        toy_txt = read_file(os.path.join(test_easyconfigs, 't', 'toy', 'toy-0.0-gompi-2018a-test.eb'))
        replace_str = '''description = """Toy C program, 100% toy. Now with `backticks'\n'''
        replace_str += '''and newlines"""'''
        toy_txt = re.sub('description = .*', replace_str, toy_txt)
        toy_file = os.path.join(self.test_prefix, 'toy-test-description.eb')
        write_file(toy_file, toy_txt)

        regex = re.compile(r"""`backticks'""")
        self.assertTrue(regex.search(toy_txt), "Pattern '%s' found in: %s" % (regex.pattern, toy_txt))
        ec_desc = EasyConfig(toy_file, validate=False)
        easyblock_desc = EB_toy(ec_desc)
        easyblock_desc.run_all_steps(False)
        pkgdir = package(easyblock_desc)
        pkgfile = os.path.join(pkgdir, 'toy-0.0-gompi-2018a-test-eb-%s.1.rpm' % EASYBUILD_VERSION)
        self.assertTrue(os.path.isfile(pkgfile))
        pkgtxt = read_file(pkgfile)
        regex_pkg = re.compile(r"""DESCRIPTION:.*`backticks'.*""")
        self.assertTrue(regex_pkg.search(pkgtxt), "Pattern '%s' not found in: %s" % (regex_pkg.pattern, pkgtxt))
        regex_pkg = re.compile(r"""DESCRIPTION:.*\nand newlines""", re.MULTILINE)
        self.assertTrue(regex_pkg.search(pkgtxt), "Pattern '%s' not found in: %s" % (regex_pkg.pattern, pkgtxt))


def suite():
    """ returns all the testcases in this module """
    return TestLoaderFiltered().loadTestsFromTestCase(PackageTest, sys.argv[1:])


if __name__ == '__main__':
    res = TextTestRunner(verbosity=1).run(suite())
    sys.exit(len(res.failures))



# #
# Copyright 2009-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #

"""
Easyconfig module that contains the default EasyConfig configuration parameters.

:author: Stijn De Weirdt (Ghent University)
:author: Dries Verdegem (Ghent University)
:author: Kenneth Hoste (Ghent University)
:author: Pieter De Baets (Ghent University)
:author: Jens Timmerman (Ghent University)
:author: Toon Willems (Ghent University)
"""
from easybuild.base import fancylogger
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import MODULECLASS_BASE


_log = fancylogger.getLogger('easyconfig.default', fname=False)

# constants for different categories of easyconfig parameters
# use tuples so we can sort them based on the numbers
HIDDEN = (-1, 'hidden')
MANDATORY = (0, 'mandatory')
CUSTOM = (1, 'easyblock-specific')
TOOLCHAIN = (2, 'toolchain')
BUILD = (3, 'build')
FILEMANAGEMENT = (4, 'file-management')
DEPENDENCIES = (5, 'dependencies')
LICENSE = (6, 'license')
EXTENSIONS = (7, 'extensions')
MODULES = (8, 'modules')
OTHER = (9, 'other')


# we use a tuple here so we can sort them based on the numbers
CATEGORY_NAMES = ['BUILD', 'CUSTOM', 'DEPENDENCIES', 'EXTENSIONS', 'FILEMANAGEMENT', 'HIDDEN',
                  'LICENSE', 'MANDATORY', 'MODULES', 'OTHER', 'TOOLCHAIN']
ALL_CATEGORIES = dict((name, eval(name)) for name in CATEGORY_NAMES)

# List of tuples. Each tuple has the following format (key, [default, help text, category])
DEFAULT_CONFIG = {
    # MANDATORY easyconfig parameters
    'description': [None, 'A short description of the software', MANDATORY],
    'homepage': [None, 'The homepage of the software', MANDATORY],
    'name': [None, "Name of software", MANDATORY],
    'toolchain': [None, 'Name and version of toolchain', MANDATORY],
    'version': [None, "Version of software", MANDATORY],
    # TODO not yet in MANDATORY_PARAMS, so not enforced (only enforced in v2)
    'software_license': [None, 'Software license', MANDATORY],
    'software_license_urls': [None, 'List of software license locations', MANDATORY],
    # TODO not yet in MANDATORY_PARAMS, so not enforced  (only enforced in v2)
    'docurls': [None, 'List of urls with documentation of the software (not necessarily on homepage)', MANDATORY],

    # TOOLCHAIN easyconfig parameters
    'onlytcmod': [False, ('Boolean/string to indicate if the toolchain should only load '
                          'the environment with module (True) or also set all other '
                          'variables (False) like compiler CC etc (if string: comma '
                          'separated list of variables that will be ignored).'), TOOLCHAIN],
    'toolchainopts': [None, 'Extra options for compilers', TOOLCHAIN],

    # BUILD easyconfig parameters
    'banned_linked_shared_libs': [[], "List of shared libraries (names, file names, or paths) which are not allowed "
                                      "to be linked in any installed binary/library", BUILD],
    'bitbucket_account': ['%(namelower)s', "Bitbucket account name to be used to resolve template values in source"
                                           " URLs", BUILD],
    'buildopts': ['', 'Extra options passed to make step (default already has -j X)', BUILD],
    'checksums': [[], "Checksums for sources and patches", BUILD],
    'configopts': ['', 'Extra options passed to configure (default already has --prefix)', BUILD],
    'cuda_compute_capabilities': [[], "List of CUDA compute capabilities to build with (if supported)", BUILD],
    'download_instructions': ['', "Specify steps to aquire necessary file, if obtaining it is difficult", BUILD],
    'easyblock': [None, "EasyBlock to use for building; if set to None, an easyblock is selected "
                        "based on the software name", BUILD],
    'easybuild_version': [None, "EasyBuild-version this spec-file was written for", BUILD],
    'enhance_sanity_check': [False, "Indicate that additional sanity check commands & paths should enhance "
                             "the existin sanity check, not replace it", BUILD],
    'fix_bash_shebang_for': [None, "List of files for which Bash shebang should be fixed "
                                   "to '#!/usr/bin/env bash' (glob patterns supported)", BUILD],
    'fix_perl_shebang_for': [None, "List of files for which Perl shebang should be fixed "
                                   "to '#!/usr/bin/env perl' (glob patterns supported)", BUILD],
    'fix_python_shebang_for': [None, "List of files for which Python shebang should be fixed "
                                     "to '#!/usr/bin/env python' (glob patterns supported)", BUILD],
    'github_account': ['%(namelower)s', "GitHub account name to be used to resolve template values in source URLs",
                       BUILD],
    'hidden': [False, "Install module file as 'hidden' by prefixing its version with '.'", BUILD],
    'installopts': ['', 'Extra options for installation', BUILD],
    'maxparallel': [None, 'Max degree of parallelism', BUILD],
    'parallel': [None, ('Degree of parallelism for e.g. make (default: based on the number of '
                        'cores, active cpuset and restrictions in ulimit)'), BUILD],
    'patches': [[], "List of patches to apply", BUILD],
    'prebuildopts': ['', 'Extra options pre-passed to build command.', BUILD],
    'preconfigopts': ['', 'Extra options pre-passed to configure.', BUILD],
    'preinstallopts': ['', 'Extra prefix options for installation.', BUILD],
    'pretestopts': ['', 'Extra prefix options for test.', BUILD],
    'postinstallcmds': [[], 'Commands to run after the install step.', BUILD],
    'postinstallpatches': [[], 'Patch files to apply after running the install step.', BUILD],
    'required_linked_shared_libs': [[], "List of shared libraries (names, file names, or paths) which must be "
                                        "linked in all installed binaries/libraries", BUILD],
    'runtest': [None, ('Indicates if a test should be run after make; should specify argument '
                       'after make (for e.g.,"test" for make test)'), BUILD],
    'bin_lib_subdirs': [[], "List of subdirectories for binaries and libraries, which is used during sanity check "
                            "to check RPATH linking and banned/required libraries", BUILD],
    'sanity_check_commands': [[], ("format: [(name, options)] e.g. [('gzip','-h')]. "
                                   "Using a non-tuple is equivalent to (name, '-h')"), BUILD],
    'sanity_check_paths': [{}, ("List of files and directories to check "
                                "(format: {'files':<list>, 'dirs':<list>})"), BUILD],
    'skip': [False, "Skip existing software", BUILD],
    'skipsteps': [[], "Skip these steps", BUILD],
    'source_urls': [[], "List of URLs for source files", BUILD],
    'sources': [[], "List of source files", BUILD],
    'stop': [None, 'Keyword to halt the build process after a certain step.', BUILD],
    'testopts': ['', 'Extra options for test.', BUILD],
    'tests': [[], ("List of test-scripts to run after install. A test script should return a "
                   "non-zero exit status to fail"), BUILD],
    'unpack_options': ['', "Extra options for unpacking source", BUILD],
    'unwanted_env_vars': [[], "List of environment variables that shouldn't be set during build", BUILD],
    'versionprefix': ['', ('Additional prefix for software version '
                           '(placed before version and toolchain name)'), BUILD],
    'versionsuffix': ['', 'Additional suffix for software version (placed after toolchain name)', BUILD],

    # FILEMANAGEMENT easyconfig parameters
    'buildininstalldir': [False, ('Boolean to build (True) or not build (False) in the installation directory'),
                          FILEMANAGEMENT],
    'cleanupoldbuild': [True, ('Boolean to remove (True) or backup (False) the previous build '
                               'directory with identical name or not.'), FILEMANAGEMENT],
    'cleanupoldinstall': [True, ('Boolean to remove (True) or backup (False) the previous install '
                                 'directory with identical name or not.'), FILEMANAGEMENT],
    'dontcreateinstalldir': [False, ('Boolean to create (False) or not create (True) the install directory'),
                             FILEMANAGEMENT],
    'keeppreviousinstall': [False, ('Boolean to keep the previous installation with identical '
                                    'name. Experts only!'), FILEMANAGEMENT],
    'keepsymlinks': [False, ('Boolean to determine whether symlinks are to be kept during copying '
                             'or if the content of the files pointed to should be copied'),
                     FILEMANAGEMENT],
    'start_dir': [None, ('Path to start the make in. If the path is absolute, use that path. '
                         'If not, this is added to the guessed path.'), FILEMANAGEMENT],

    # DEPENDENCIES easyconfig parameters
    'allow_system_deps': [[], "Allow listed system dependencies (format: (<name>, <version>))", DEPENDENCIES],
    'builddependencies': [[], "List of build dependencies", DEPENDENCIES],
    'dependencies': [[], "List of dependencies", DEPENDENCIES],
    'hiddendependencies': [[], "List of dependencies available as hidden modules", DEPENDENCIES],
    'multi_deps': [{}, "Dict of lists of dependency versions over which to iterate", DEPENDENCIES],
    'multi_deps_load_default': [True, "Load module for first version listed in multi_deps by default", DEPENDENCIES],
    'osdependencies': [[], "OS dependencies that should be present on the system", DEPENDENCIES],
    'moddependpaths': [None, "Absolute path(s) to prepend to MODULEPATH before loading dependencies", DEPENDENCIES],

    # LICENSE easyconfig parameters
    'accept_eula': [False, "Accepted End User License Agreement (EULA) for this software", LICENSE],
    'group': [None, "Name of the user group for which the software should be available; "
                    "format: string or 2-tuple with group name + custom error for users outside group", LICENSE],
    'key': [None, 'Key for installing software', LICENSE],
    'license_file': [None, 'License file for software', LICENSE],
    'license_server': [None, 'License server for software', LICENSE],
    'license_server_port': [None, 'Port for license server', LICENSE],

    # EXTENSIONS easyconfig parameters
    'exts_download_dep_fail': [False, "Fail if downloaded dependencies are detected for extensions", EXTENSIONS],
    'exts_classmap': [{}, "Map of extension name to class for handling build and installation.", EXTENSIONS],
    'exts_defaultclass': [None, "List of module for and name of the default extension class", EXTENSIONS],
    'exts_default_options': [{}, "List of default options for extensions", EXTENSIONS],
    'exts_filter': [None, ("Extension filter details: template for cmd and input to cmd "
                           "(templates for ext_name, ext_version and src)."), EXTENSIONS],
    'exts_list': [[], 'List with extensions added to the base installation', EXTENSIONS],

    # MODULES easyconfig parameters
    'allow_prepend_abs_path': [False, "Allow specifying absolute paths to prepend in modextrapaths", MODULES],
    'include_modpath_extensions': [True, "Include $MODULEPATH extensions specified by module naming scheme.", MODULES],
    'modaliases': [{}, "Aliases to be defined in module file", MODULES],
    'modextrapaths': [{}, "Extra paths to be prepended in module file", MODULES],
    'modextravars': [{}, "Extra environment variables to be added to module file", MODULES],
    'modloadmsg': [{}, "Message that should be printed when generated module is loaded", MODULES],
    'modluafooter': ["", "Footer to include in generated module file (Lua syntax)", MODULES],
    'modaltsoftname': [None, "Module name to use (rather than using software name", MODULES],
    'modtclfooter': ["", "Footer to include in generated module file (Tcl syntax)", MODULES],
    'moduleclass': [MODULECLASS_BASE, 'Module class to be used for this software', MODULES],
    'moduleforceunload': [False, 'Force unload of all modules when loading the extension', MODULES],
    'moduleloadnoconflict': [False, "Don't check for conflicts, unload other versions instead ", MODULES],
    'module_depends_on': [False, 'Use depends_on (Lmod 7.6.1+) for dependencies in generated module '
                          '(implies recursive unloading of modules).', MODULES],
    'recursive_module_unload': [None, "Recursive unload of all dependencies when unloading module "
                                      "(True/False to hard enable/disable; None implies honoring "
                                      "the --recursive-module-unload EasyBuild configuration setting",
                                MODULES],

    # MODULES documentation easyconfig parameters
    #    (docurls is part of MANDATORY)
    'citing': [None, "Free-form text that describes how the software should be cited in publications", MODULES],
    'docpaths': [None, "List of paths for documentation relative to installation directory", MODULES],
    'examples': [None, "Free-form text with examples on using the software", MODULES],
    'site_contacts': [None, "String/list of strings with site contacts for the software", MODULES],
    'upstream_contacts': [None, "String/list of strings with upstream contact addresses "
                          "(e.g., support e-mail, mailing list, bugtracker)", MODULES],
    'usage': [None, "Usage instructions for the software", MODULES],
    'whatis': [None, "List of brief (one line) description entries for the software", MODULES],

    # OTHER easyconfig parameters
    # 'block' must be a known easyconfig parameter in case strict local variable naming is enabled;
    # see also retrieve_blocks_in_spec function
    'block': [None, "List of other 'block' sections on which this block depends "
                    "(only relevant in easyconfigs with subblocks)", OTHER],
    'buildstats': [None, "A list of dicts with build statistics", OTHER],
    'deprecated': [False, "String specifying reason why this easyconfig file is deprecated "
                          "and will be archived in the next major release of EasyBuild", OTHER],
}


def sorted_categories():
    """
    returns the categories in the correct order
    """
    categories = list(ALL_CATEGORIES.values())
    categories.sort(key=lambda c: c[0])
    return categories


def get_easyconfig_parameter_default(param):
    """Get default value for given easyconfig parameter."""
    if param not in DEFAULT_CONFIG:
        raise EasyBuildError("Unknown easyconfig parameter: %s (known: %s)", param, sorted(DEFAULT_CONFIG.keys()))
    else:
        _log.debug("Returning default value for easyconfig parameter %s: %s" % (param, DEFAULT_CONFIG[param][0]))
        return DEFAULT_CONFIG[param][0]


def is_easyconfig_parameter_default_value(param, value):
    """Return True if the parameters is one of the default ones and the value equals its default value"""
    return param in DEFAULT_CONFIG and get_easyconfig_parameter_default(param) == value



# #
# Copyright 2009-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Support for generating docker container recipes and creating container images

:author Mohamed Abidi (Bright Computing)
"""
import os
import tempfile

from easybuild.tools.build_log import EasyBuildError, print_msg
from easybuild.tools.config import DOCKER_BASE_IMAGE_CENTOS, DOCKER_BASE_IMAGE_UBUNTU
from easybuild.tools.containers.base import ContainerGenerator
from easybuild.tools.containers.utils import det_os_deps
from easybuild.tools.filetools import remove_dir
from easybuild.tools.module_naming_scheme.easybuild_mns import EasyBuildMNS
from easybuild.tools.run import run_cmd


DOCKER_TMPL_HEADER = """\
FROM %(container_config)s
LABEL maintainer=easybuild@lists.ugent.be
"""

DOCKER_INSTALL_EASYBUILD = """\
RUN pip3 install -U pip setuptools && \\
    hash -r pip3&& \\
    pip3 install -U easybuild

RUN mkdir /app && \\
    mkdir /scratch && \\
    mkdir /scratch/tmp && \\
    useradd -m -s /bin/bash easybuild && \\
    chown easybuild:easybuild -R /app && \\
    chown easybuild:easybuild -R /scratch
"""

DOCKER_TMPL_FOOTER = """\
USER easybuild

RUN set -x && \\
    . /usr/share/lmod/lmod/init/sh && \\
    eb --robot %(eb_opts)s --installpath=/app/ --prefix=/scratch --tmpdir=/scratch/tmp

RUN touch ${HOME}/.bashrc && \\
    echo '' >> ${HOME}/.bashrc && \\
    echo '# Added by easybuild docker packaging' >> ${HOME}/.bashrc && \\
    echo 'source /usr/share/lmod/lmod/init/bash' >> ${HOME}/.bashrc && \\
    echo 'module use %(init_modulepath)s' >> ${HOME}/.bashrc && \\
    echo 'module load %(mod_names)s' >> ${HOME}/.bashrc

CMD ["/bin/bash", "-l"]
"""

DOCKER_UBUNTU2004_INSTALL_DEPS = """\
RUN apt-get update && \\
    DEBIAN_FRONTEND=noninteractive apt-get install -y python3 python3-pip lmod \\
        curl wget git bzip2 gzip tar zip unzip xz-utils \\
        patch automake git debianutils \\
        g++ libdata-dump-perl libthread-queue-any-perl libssl-dev

RUN OS_DEPS='%(os_deps)s' && \\
    for dep in ${OS_DEPS}; do apt-get -qq install ${dep} || true; done
"""

DOCKER_CENTOS7_INSTALL_DEPS = """\
RUN yum install -y epel-release && \\
    yum install -y python3 python3-pip Lmod curl wget git \\
        bzip2 gzip tar zip unzip xz \\
        patch make git which \\
        gcc-c++ perl-Data-Dumper perl-Thread-Queue openssl-dev

RUN OS_DEPS='%(os_deps)s' && \\
    test -n "${OS_DEPS}" && \\
    yum --skip-broken install -y "${OS_DEPS}" || true
"""

DOCKER_OS_INSTALL_DEPS_TMPLS = {
    DOCKER_BASE_IMAGE_UBUNTU: DOCKER_UBUNTU2004_INSTALL_DEPS,
    DOCKER_BASE_IMAGE_CENTOS: DOCKER_CENTOS7_INSTALL_DEPS,
}


class DockerContainer(ContainerGenerator):

    TOOLS = {'docker': None, 'sudo': None}

    RECIPE_FILE_NAME = 'Dockerfile'

    def resolve_template(self):
        """Return template container recipe."""

        if self.container_template_recipe:
            raise EasyBuildError("--container-template-recipe is not supported yet for Docker container images!")

        if self.container_config:
            return '\n\n'.join([
                DOCKER_TMPL_HEADER % {'container_config': self.container_config},
                DOCKER_OS_INSTALL_DEPS_TMPLS[self.container_config],
                DOCKER_INSTALL_EASYBUILD,
                DOCKER_TMPL_FOOTER,
            ])
        else:
            raise EasyBuildError("--container--config is required for Docker container images!")

    def resolve_template_data(self):
        """Return template data for container recipe."""

        os_deps = det_os_deps(self.easyconfigs)

        ec = self.easyconfigs[-1]['ec']

        # We are using the default MNS inside the container
        docker_mns = EasyBuildMNS()

        init_modulepath = os.path.join("/app/modules/all", *docker_mns.det_init_modulepaths(ec))

        mod_names = [docker_mns.det_full_module_name(e['ec']) for e in self.easyconfigs]

        eb_opts = [os.path.basename(e['spec']) for e in self.easyconfigs]

        return {
            'os_deps': ' '.join(os_deps),
            'eb_opts': ' '.join(eb_opts),
            'init_modulepath': init_modulepath,
            'mod_names': ' '.join(mod_names),
        }

    def validate(self):
        """Perform validation of specified container configuration."""
        if self.container_config not in DOCKER_OS_INSTALL_DEPS_TMPLS.keys():
            raise EasyBuildError("Unsupported container config '%s'" % self.container_config)
        super(DockerContainer, self).validate()

    def build_image(self, dockerfile):
        ec = self.easyconfigs[-1]['ec']

        module_name = self.mns.det_full_module_name(ec)

        tempdir = tempfile.mkdtemp(prefix='easybuild-docker')
        container_name = self.img_name or "%s:latest" % module_name.replace('/', '-')
        docker_cmd = ' '.join(['sudo', 'docker', 'build', '-f', dockerfile, '-t', container_name, '.'])

        print_msg("Running '%s', you may need to enter your 'sudo' password..." % docker_cmd)
        run_cmd(docker_cmd, path=tempdir, stream_output=True)
        print_msg("Docker image created at %s" % container_name, log=self.log)

        remove_dir(tempdir)



# #
# Copyright 2009-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Generating module files.

:author: Stijn De Weirdt (Ghent University)
:author: Dries Verdegem (Ghent University)
:author: Kenneth Hoste (Ghent University)
:author: Pieter De Baets (Ghent University)
:author: Jens Timmerman (Ghent University)
:author: Fotis Georgatos (Uni.Lu, NTUA)
:author: Damian Alvarez (Forschungszentrum Juelich GmbH)
"""
import copy
import os
import re
import tempfile
from contextlib import contextmanager
from distutils.version import LooseVersion
from textwrap import wrap

from easybuild.base import fancylogger
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option, get_module_syntax, install_path
from easybuild.tools.filetools import convert_name, mkdir, read_file, remove_file, resolve_path, symlink, write_file
from easybuild.tools.modules import ROOT_ENV_VAR_NAME_PREFIX, EnvironmentModulesC, Lmod, modules_tool
from easybuild.tools.py2vs3 import string_type
from easybuild.tools.utilities import get_subclasses, quote_str


_log = fancylogger.getLogger('module_generator', fname=False)


def avail_module_generators():
    """
    Return all known module syntaxes.
    """
    return dict([(k.SYNTAX, k) for k in get_subclasses(ModuleGenerator)])


def module_generator(app, fake=False):
    """
    Return ModuleGenerator instance that matches the selected module file syntax to be used
    """
    module_syntax = get_module_syntax()
    available_mod_gens = avail_module_generators()

    if module_syntax not in available_mod_gens:
        raise EasyBuildError("No module generator available for specified syntax '%s' (available: %s)",
                             module_syntax, available_mod_gens)

    module_generator_class = available_mod_gens[module_syntax]
    return module_generator_class(app, fake=fake)


def module_load_regex(modfilepath):
    """
    Return the correct (compiled) regex to extract dependencies, depending on the module file type (Lua vs Tcl)
    """
    if modfilepath.endswith(ModuleGeneratorLua.MODULE_FILE_EXTENSION):
        regex = ModuleGeneratorLua.LOAD_REGEX
    else:
        regex = ModuleGeneratorTcl.LOAD_REGEX
    return re.compile(regex, re.M)


def dependencies_for(mod_name, modtool, depth=None):
    """
    Obtain a list of dependencies for the given module, determined recursively, up to a specified depth (optionally)
    :param depth: recursion depth (default is None, which corresponds to infinite recursion depth)
    """
    mod_filepath = modtool.modulefile_path(mod_name)
    modtxt = read_file(mod_filepath)
    loadregex = module_load_regex(mod_filepath)
    mods = loadregex.findall(modtxt)

    if depth is None or depth > 0:
        if depth and depth > 0:
            depth = depth - 1
        # recursively determine dependencies for these dependency modules, until depth is non-positive
        moddeps = [dependencies_for(mod, modtool, depth=depth) for mod in mods]
    else:
        # ignore any deeper dependencies
        moddeps = []

    # add dependencies of dependency modules only if they're not there yet
    for moddepdeps in moddeps:
        for dep in moddepdeps:
            if dep not in mods:
                mods.append(dep)

    return mods


class ModuleGenerator(object):
    """
    Class for generating module files.
    """
    SYNTAX = None

    # chars we want to escape in the generated modulefiles
    CHARS_TO_ESCAPE = None

    MODULE_FILE_EXTENSION = None
    MODULE_SHEBANG = None
    DOT_MODULERC = '.modulerc'

    # a single level of indentation
    INDENTATION = ' ' * 4

    def __init__(self, application, fake=False):
        """ModuleGenerator constructor."""
        self.app = application
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)
        self.fake_mod_path = tempfile.mkdtemp()

        self.modules_tool = modules_tool()
        self.added_paths_per_key = None

    @contextmanager
    def start_module_creation(self):
        """
        Prepares creating a module and returns the file header (shebang) if any including the newline

        Meant to be used in a with statement:
            with generator.start_module_creation() as txt:
                # Write txt
        """
        if self.added_paths_per_key is not None:
            raise EasyBuildError('Module creation already in process. '
                                 'You cannot create multiple modules at the same time!')
        # Mapping of keys/env vars to paths already added
        self.added_paths_per_key = dict()
        txt = self.MODULE_SHEBANG
        if txt:
            txt += '\n'
        try:
            yield txt
        finally:
            self.added_paths_per_key = None

    def create_symlinks(self, mod_symlink_paths, fake=False):
        """Create moduleclass symlink(s) to actual module file."""
        mod_filepath = self.get_module_filepath(fake=fake)
        class_mod_files = [self.get_module_filepath(fake=fake, mod_path_suffix=p) for p in mod_symlink_paths]
        try:
            for class_mod_file in class_mod_files:
                # remove symlink if its there (even if it's broken)
                if os.path.lexists(class_mod_file):
                    self.log.debug("Removing existing symlink %s", class_mod_file)
                    os.remove(class_mod_file)

                mkdir(os.path.dirname(class_mod_file), parents=True)
                os.symlink(mod_filepath, class_mod_file)

        except OSError as err:
            raise EasyBuildError("Failed to create symlinks from %s to %s: %s", class_mod_files, mod_filepath, err)

    def define_env_var(self, env_var):
        """
        Determine whether environment variable with specified name should be defined or not.

        :param env_var: name of environment variable to check
        """
        return env_var not in (build_option('filter_env_vars') or [])

    def get_module_filepath(self, fake=False, mod_path_suffix=None):
        """Return path to module file."""
        mod_path = self.get_modules_path(fake=fake, mod_path_suffix=mod_path_suffix)
        full_mod_name = self.app.full_mod_name + self.MODULE_FILE_EXTENSION
        return os.path.join(mod_path, full_mod_name)

    def get_modules_path(self, fake=False, mod_path_suffix=None):
        """Return path to directory where module files should be generated in."""
        mod_path = install_path('mod')
        if fake:
            self.log.debug("Fake mode: using %s (instead of %s)" % (self.fake_mod_path, mod_path))
            mod_path = self.fake_mod_path

        if mod_path_suffix is None:
            mod_path_suffix = build_option('suffix_modules_path')

        return os.path.join(mod_path, mod_path_suffix)

    def _filter_paths(self, key, paths):
        """Filter out paths already added to key and return the remaining ones"""
        if self.added_paths_per_key is None:
            # For compatibility this is only a warning for now and we don't filter any paths
            print_warning('Module creation has not been started. Call start_module_creation first!')
            return paths

        added_paths = self.added_paths_per_key.setdefault(key, set())
        # paths can be a string
        if isinstance(paths, string_type):
            if paths in added_paths:
                filtered_paths = None
            else:
                added_paths.add(paths)
                filtered_paths = paths
        else:
            # Coerce any iterable/generator into a list
            if not isinstance(paths, list):
                paths = list(paths)
            filtered_paths = [x for x in paths if x not in added_paths and not added_paths.add(x)]
        if filtered_paths != paths:
            removed_paths = paths if filtered_paths is None else [x for x in paths if x not in filtered_paths]
            print_warning("Suppressed adding the following path(s) to $%s of the module as they were already added: %s",
                          key, removed_paths,
                          log=self.log)
            if not filtered_paths:
                filtered_paths = None
        return filtered_paths

    def append_paths(self, key, paths, allow_abs=False, expand_relpaths=True):
        """
        Generate append-path statements for the given list of paths.

        :param key: environment variable to append paths to
        :param paths: list of paths to append
        :param allow_abs: allow providing of absolute paths
        :param expand_relpaths: expand relative paths into absolute paths (by prefixing install dir)
        """
        paths = self._filter_paths(key, paths)
        if paths is None:
            return ''
        return self.update_paths(key, paths, prepend=False, allow_abs=allow_abs, expand_relpaths=expand_relpaths)

    def prepend_paths(self, key, paths, allow_abs=False, expand_relpaths=True):
        """
        Generate prepend-path statements for the given list of paths.

        :param key: environment variable to append paths to
        :param paths: list of paths to append
        :param allow_abs: allow providing of absolute paths
        :param expand_relpaths: expand relative paths into absolute paths (by prefixing install dir)
        """
        paths = self._filter_paths(key, paths)
        if paths is None:
            return ''
        return self.update_paths(key, paths, prepend=True, allow_abs=allow_abs, expand_relpaths=expand_relpaths)

    def _modulerc_check_module_version(self, module_version):
        """
        Check value type & contents of specified module-version spec.

        :param module_version: specs for module-version statement (dict with 'modname', 'sym_version' & 'version' keys)
        :return: True if spec is OK
        """
        res = False
        if module_version:
            if isinstance(module_version, dict):
                expected_keys = ['modname', 'sym_version', 'version']
                if sorted(module_version.keys()) == expected_keys:
                    res = True
                else:
                    raise EasyBuildError("Incorrect module_version spec, expected keys: %s", expected_keys)
            else:
                raise EasyBuildError("Incorrect module_version value type: %s", type(module_version))

        return res

    def _write_modulerc_file(self, modulerc_path, modulerc_txt, wrapped_mod_name=None):
        """
        Write modulerc file with specified contents.

        :param modulerc_path: location of .modulerc file to write
        :param modulerc_txt: contents of .modulerc file
        :param wrapped_mod_name: name of module file for which a wrapper is defined in the .modulerc file (if any)
        """
        # Lmod 6.x requires that module being wrapped is in same location as .modulerc file...
        if wrapped_mod_name is not None:
            if isinstance(self.modules_tool, Lmod) and LooseVersion(self.modules_tool.version) < LooseVersion('7.0'):
                mod_dir = os.path.dirname(modulerc_path)

                # need to consider existing module file in both Tcl (no extension) & Lua (.lua extension) syntax...
                wrapped_mod_fp = os.path.join(mod_dir, os.path.basename(wrapped_mod_name))
                wrapped_mod_exists = os.path.exists(wrapped_mod_fp)
                if not wrapped_mod_exists and self.MODULE_FILE_EXTENSION:
                    wrapped_mod_exists = os.path.exists(wrapped_mod_fp + self.MODULE_FILE_EXTENSION)

                if not wrapped_mod_exists:
                    error_msg = "Expected module file %s not found; " % wrapped_mod_fp
                    error_msg += "Lmod 6.x requires that .modulerc and wrapped module file are in same directory!"
                    raise EasyBuildError(error_msg)

        if os.path.exists(modulerc_path):
            curr_modulerc = read_file(modulerc_path)

            # get rid of Tcl shebang line if modulerc file already exists and already contains Tcl shebang line
            tcl_shebang = ModuleGeneratorTcl.MODULE_SHEBANG
            if modulerc_txt.startswith(tcl_shebang) and curr_modulerc.startswith(tcl_shebang):
                modulerc_txt = '\n'.join(modulerc_txt.split('\n')[1:])

            # check whether specified contents is already contained in current modulerc file;
            # if so, we don't need to update the existing modulerc at all...
            # if it's not, we need to append to existing modulerc file
            if modulerc_txt.strip() not in curr_modulerc:

                # if current contents doesn't end with a newline, prefix text being appended with a newline
                if not curr_modulerc.endswith('\n'):
                    modulerc_txt = '\n' + modulerc_txt

                write_file(modulerc_path, modulerc_txt, append=True, backup=True)
        else:
            write_file(modulerc_path, modulerc_txt)

    def modulerc(self, module_version=None, filepath=None, modulerc_txt=None):
        """
        Generate contents of .modulerc file, in Tcl syntax (compatible with all module tools, incl. Lmod).
        If 'filepath' is specified, the .modulerc file will be written as well.

        :param module_version: specs for module-version statement (dict with 'modname', 'sym_version' & 'version' keys)
        :param filepath: location where .modulerc file should be written to
        :param modulerc_txt: contents of .modulerc to use
        :return: contents of .modulerc file
        """
        if modulerc_txt is None:

            self.log.info("Generating .modulerc contents in Tcl syntax (args: module_version: %s", module_version)
            modulerc = [ModuleGeneratorTcl.MODULE_SHEBANG]

            if self._modulerc_check_module_version(module_version):

                module_version_statement = "module-version %(modname)s %(sym_version)s"

                # for Environment Modules we need to guard the module-version statement,
                # to avoid "Duplicate version symbol" warning messages where EasyBuild trips over,
                # which occur because the .modulerc is parsed twice
                # "module-info version <arg>" returns its argument if that argument is not a symbolic version (yet),
                # and returns the corresponding real version in case the argument is an existing symbolic version
                # cfr. https://sourceforge.net/p/modules/mailman/message/33399425/
                if self.modules_tool.__class__ == EnvironmentModulesC:

                    keys = ['modname', 'sym_version', 'version']
                    modname, sym_version, version = [module_version[key] for key in keys]

                    # determine module name with symbolic version
                    if version in modname:
                        # take a copy so we don't modify original value
                        module_version = copy.copy(module_version)
                        module_version['sym_modname'] = modname.replace(version, sym_version)
                    else:
                        raise EasyBuildError("Version '%s' does not appear in module name '%s'", version, modname)

                    module_version_statement = '\n'.join([
                        'if {"%(sym_modname)s" eq [module-info version %(sym_modname)s]} {',
                        ' ' * 4 + module_version_statement,
                        "}",
                    ])

                modulerc.append(module_version_statement % module_version)

            modulerc_txt = '\n'.join(modulerc)

        if filepath:
            self.log.info("Writing %s with contents:\n%s", filepath, modulerc_txt)
            self._write_modulerc_file(filepath, modulerc_txt, wrapped_mod_name=module_version['modname'])

        return modulerc_txt

    def is_loaded(self, mod_names):
        """
        Generate (list of) expression(s) to check whether specified module(s) is (are) loaded.

        :param mod_names: (list of) module name(s) to check load status for
        """
        if isinstance(mod_names, string_type):
            res = self.IS_LOADED_TEMPLATE % mod_names
        else:
            res = [self.IS_LOADED_TEMPLATE % m for m in mod_names]

        return res

    def det_installdir(self, modfile):
        """
        Determine installation directory used by given module file
        """
        res = None

        modtxt = read_file(modfile)
        root_regex = re.compile(self.INSTALLDIR_REGEX, re.M)
        match = root_regex.search(modtxt)
        if match:
            res = match.group('installdir')

        return res

    # From this point on just not implemented methods

    def check_group(self, group, error_msg=None):
        """
        Generate a check of the software group and the current user, and refuse to load the module if the user don't
        belong to the group

        :param group: string with the group name
        :param error_msg: error message to print for users outside that group
        """
        raise NotImplementedError

    def comment(self, msg):
        """Return given string formatted as a comment."""
        raise NotImplementedError

    def check_version(self, minimal_version_maj, minimal_version_min, minimal_version_patch='0'):
        """
        Check the minimal version of the modules tool in the module file
        :param minimal_version_maj: the major version to check
        :param minimal_version_min: the minor version to check
        :param minimal_version_patch: the patch version to check
        """
        raise NotImplementedError

    def conditional_statement(self, conditions, body, negative=False, else_body=None, indent=True,
                              cond_or=False, cond_tmpl=None):
        """
        Return formatted conditional statement, with given condition and body.

        :param conditions: (list of) string(s) containing the statement(s) for the if condition (in correct syntax)
        :param body: (multiline) string with if body (in correct syntax, without indentation)
        :param negative: boolean indicating whether the (individual) condition(s) should be negated
        :param else_body: optional body for 'else' part
        :param indent: indent if/else body
        :param cond_or: combine multiple conditions using 'or' (default is to combine with 'and')
        :param cond_tmpl: template for condition expression (default: '%s')
        """
        raise NotImplementedError

    def get_description(self, conflict=True):
        """
        Generate a description.
        """
        raise NotImplementedError

    def getenv_cmd(self, envvar, default=None):
        """
        Return module-syntax specific code to get value of specific environment variable.
        """
        raise NotImplementedError

    def load_module(self, mod_name, recursive_unload=False, depends_on=False, unload_modules=None, multi_dep_mods=None):
        """
        Generate load statement for specified module.

        :param mod_name: name of module to generate load statement for
        :param recursive_unload: boolean indicating whether the 'load' statement should be reverted on unload
        :param depends_on: use depends_on statements rather than (guarded) load statements
        :param unload_modules: name(s) of module to unload first
        :param multi_dep_mods: list of module names in multi_deps context, to use for guarding load statement
        """
        raise NotImplementedError

    def msg_on_load(self, msg):
        """
        Add a message that should be printed when loading the module.
        """
        raise NotImplementedError

    def set_alias(self, key, value):
        """
        Generate set-alias statement in modulefile for the given key/value pair.
        """
        raise NotImplementedError

    def set_as_default(self, module_dir_path, module_version, mod_symlink_paths=None):
        """
        Set generated module as default module

        :param module_dir_path: module directory path, e.g. $HOME/easybuild/modules/all/Bison
        :param module_version: module version, e.g. 3.0.4
        :param mod_symlink_paths: list of paths in which symlinks to module files must be created
        """
        raise NotImplementedError

    def set_environment(self, key, value, relpath=False):
        """
        Generate a quoted setenv statement for the given key/value pair.

        :param key: name of environment variable to define
        :param value: value to define environment variable with
        :param relpath: value is path relative to installation prefix
        """
        raise NotImplementedError

    def swap_module(self, mod_name_out, mod_name_in, guarded=True):
        """
        Generate swap statement for specified module names.

        :param mod_name_out: name of module to unload (swap out)
        :param mod_name_in: name of module to load (swap in)
        :param guarded: guard 'swap' statement, fall back to 'load' if module being swapped out is not loaded
        """
        raise NotImplementedError

    def unload_module(self, mod_name):
        """
        Generate unload statement for specified module.

        :param mod_name: name of module to generate unload statement for
        """
        raise NotImplementedError

    def update_paths(self, key, paths, prepend=True, allow_abs=False, expand_relpaths=True):
        """
        Generate prepend-path or append-path statements for the given list of paths.

        :param key: environment variable to prepend/append paths to
        :param paths: list of paths to prepend
        :param prepend: whether to prepend (True) or append (False) paths
        :param allow_abs: allow providing of absolute paths
        :param expand_relpaths: expand relative paths into absolute paths (by prefixing install dir)
        """
        raise NotImplementedError

    def _det_user_modpath_common(self, user_modpath):
        """
        Helper function for det_user_modpath.
        """
        # Check for occurences of {RUNTIME_ENV::SOME_ENV_VAR}
        # SOME_ENV_VAR will be expanded at module load time.
        runtime_env_re = re.compile(r'{RUNTIME_ENV::(\w+)}')
        sub_paths = []
        expanded_user_modpath = []
        for sub_path in re.split(os.path.sep, user_modpath):
            matched_re = runtime_env_re.match(sub_path)
            if matched_re:
                if sub_paths:
                    path = quote_str(os.path.join(*sub_paths))
                    expanded_user_modpath.extend([path])
                    sub_paths = []
                expanded_user_modpath.extend([self.getenv_cmd(matched_re.group(1))])
            else:
                sub_paths.append(sub_path)
        if sub_paths:
            expanded_user_modpath.extend([quote_str(os.path.join(*sub_paths))])

        # if a mod_path_suffix is being used, we should respect it
        mod_path_suffix = build_option('suffix_modules_path')
        if mod_path_suffix:
            expanded_user_modpath.extend([quote_str(mod_path_suffix)])

        return expanded_user_modpath

    def det_user_modpath(self, user_modpath):
        """
        Determine user-specific modules subdirectory, to be used in 'use' statements
        (cfr. implementation of use() method).
        """
        raise NotImplementedError

    def use(self, paths, prefix=None, guarded=False, user_modpath=None):
        """
        Generate module use statements for given list of module paths.
        :param paths: list of module path extensions to generate use statements for; paths will be quoted
        :param prefix: optional path prefix; not quoted, i.e., can be a statement
        :param guarded: use statements will be guarded to only apply if path exists
        :param user_modpath: user-specific modules subdirectory to include in use statements
        """
        raise NotImplementedError

    def _generate_extension_list(self):
        """
        Generate a string with a comma-separated list of extensions.
        """
        # We need only name and version, so don't resolve templates
        exts_list = self.app.cfg.get_ref('exts_list')
        extensions = ', '.join(sorted(['-'.join(ext[:2]) for ext in exts_list], key=str.lower))

        return extensions

    def _generate_extensions_list(self):
        """
        Generate a list of all extensions in name/version format
        """
        exts_list = self.app.cfg['exts_list']
        # the format is extension_name/extension_version
        exts_ver_list = []
        for ext in exts_list:
            if isinstance(ext, tuple):
                exts_ver_list.append('%s/%s' % (ext[0], ext[1]))
            elif isinstance(ext, string_type):
                exts_ver_list.append(ext)

        return sorted(exts_ver_list, key=str.lower)

    def _generate_help_text(self):
        """
        Generate syntax-independent help text used for `module help`.
        """

        # General package description (mandatory)
        lines = self._generate_section('Description', self.app.cfg['description'], strip=True)

        # Package usage instructions (optional)
        lines.extend(self._generate_section('Usage', self.app.cfg['usage'], strip=True))

        # Examples (optional)
        lines.extend(self._generate_section('Examples', self.app.cfg['examples'], strip=True))

        # Citing (optional)
        lines.extend(self._generate_section('Citing', self.app.cfg['citing'], strip=True))

        # Additional information: homepage + (if available) doc paths/urls, upstream/site contact
        lines.extend(self._generate_section("More information", " - Homepage: %s" % self.app.cfg['homepage']))

        docpaths = self.app.cfg['docpaths'] or []
        docurls = self.app.cfg['docurls'] or []
        if docpaths or docurls:
            root_envvar = ROOT_ENV_VAR_NAME_PREFIX + convert_name(self.app.name, upper=True)
            lines.extend([" - Documentation:"])
            lines.extend(["    - $%s/%s" % (root_envvar, path) for path in docpaths])
            lines.extend(["    - %s" % url for url in docurls])

        for contacts_type in ['upstream', 'site']:
            contacts = self.app.cfg['%s_contacts' % contacts_type]
            if contacts:
                if isinstance(contacts, list):
                    lines.append(" - %s contacts:" % contacts_type.capitalize())
                    lines.extend(["    - %s" % contact for contact in contacts])
                else:
                    lines.append(" - %s contact: %s" % (contacts_type.capitalize(), contacts))

        # Multi deps (if any)
        multi_deps = self._generate_multi_deps_list()
        if multi_deps:
            compatible_modules_txt = '\n'.join([
                "This module is compatible with the following modules, one of each line is required:",
            ] + ['* %s' % d for d in multi_deps])
            lines.extend(self._generate_section("Compatible modules", compatible_modules_txt))

        # Extensions (if any)
        extensions = self._generate_extension_list()
        lines.extend(self._generate_section("Included extensions", '\n'.join(wrap(extensions, 78))))

        return '\n'.join(lines)

    def _generate_multi_deps_list(self):
        """
        Generate a string with a comma-separated list of multi_deps.
        """
        multi_deps = []
        if self.app.cfg['multi_deps']:
            for key in sorted(self.app.cfg['multi_deps'].keys()):
                mod_list = []
                txt = ''
                vlist = self.app.cfg['multi_deps'].get(key)
                for idx in range(len(vlist)):
                    for deplist in self.app.cfg.multi_deps:
                        for dep in deplist:
                            if dep['name'] == key and dep['version'] == vlist[idx]:
                                modname = dep['short_mod_name']
                                # indicate which version is loaded by default (unless that's disabled)
                                if idx == 0 and self.app.cfg['multi_deps_load_default']:
                                    modname += ' (default)'
                                mod_list.append(modname)
                txt += ', '.join(mod_list)
                multi_deps.append(txt)

        return multi_deps

    def _generate_section(self, sec_name, sec_txt, strip=False):
        """
        Generate section with given name and contents.
        """
        res = []
        if sec_txt:
            if strip:
                sec_txt = sec_txt.strip()
            res = ['', '', sec_name, '=' * len(sec_name), sec_txt]
        return res

    def _generate_whatis_lines(self):
        """
        Generate a list of entries used for `module whatis`.
        """
        whatis = self.app.cfg['whatis']
        if whatis is None:
            # default: include 'whatis' statements with description, homepage, and extensions (if any)
            whatis = [
                "Description: %s" % self.app.cfg['description'],
                "Homepage: %s" % self.app.cfg['homepage'],
                "URL: %s" % self.app.cfg['homepage'],
            ]

            multi_deps = self._generate_multi_deps_list()
            if multi_deps:
                whatis.append("Compatible modules: %s" % ', '.join(multi_deps))

            extensions = self._generate_extension_list()
            if extensions:
                whatis.append("Extensions: %s" % extensions)

        return whatis


class ModuleGeneratorTcl(ModuleGenerator):
    """
    Class for generating Tcl module files.
    """
    SYNTAX = 'Tcl'
    MODULE_FILE_EXTENSION = ''  # no suffix for Tcl module files
    MODULE_SHEBANG = '#%Module'
    CHARS_TO_ESCAPE = ['$']

    INSTALLDIR_REGEX = r"^set root\s+(?P<installdir>.*)"
    LOAD_REGEX = r"^\s*(?:module\s+load|depends-on)\s+(\S+)"
    LOAD_TEMPLATE = "module load %(mod_name)s"
    LOAD_TEMPLATE_DEPENDS_ON = "depends-on %(mod_name)s"
    IS_LOADED_TEMPLATE = 'is-loaded %s'

    def check_group(self, group, error_msg=None):
        """
        Generate a check of the software group and the current user, and refuse to load the module if the user don't
        belong to the group

        :param group: string with the group name
        :param error_msg: error message to print for users outside that group
        """
        self.log.warning("Can't generate robust check in TCL modules for users belonging to group %s.", group)
        return ''

    def comment(self, msg):
        """Return string containing given message as a comment."""
        return "# %s\n" % msg

    def conditional_statement(self, conditions, body, negative=False, else_body=None, indent=True,
                              cond_or=False, cond_tmpl=None):
        """
        Return formatted conditional statement, with given condition and body.

        :param conditions: (list of) string(s) containing the statement(s) for the if condition (in correct syntax)
        :param body: (multiline) string with if body (in correct syntax, without indentation)
        :param negative: boolean indicating whether the (individual) condition(s) should be negated
        :param else_body: optional body for 'else' part
        :param indent: indent if/else body
        :param cond_or: combine multiple conditions using 'or' (default is to combine with 'and')
        :param cond_tmpl: template for condition expression (default: '%s')
        """
        if isinstance(conditions, string_type):
            conditions = [conditions]

        if cond_or:
            join_op = ' || '
        else:
            join_op = ' && '

        if negative:
            condition = join_op.join('![ %s ]' % c for c in conditions)
        else:
            condition = join_op.join('[ %s ]' % c for c in conditions)

        if cond_tmpl:
            condition = cond_tmpl % condition

        lines = ["if { %s } {" % condition]

        for line in body.split('\n'):
            if indent:
                line = self.INDENTATION + line
            lines.append(line)

        if else_body is None:
            lines.extend(['}', ''])
        else:
            lines.append('} else {')
            for line in else_body.split('\n'):
                if indent:
                    line = self.INDENTATION + line
                lines.append(line)
            lines.extend(['}', ''])

        return '\n'.join(lines)

    def get_description(self, conflict=True):
        """
        Generate a description.
        """
        txt = '\n'.join([
            "proc ModulesHelp { } {",
            "    puts stderr {%s" % re.sub(r'([{}\[\]])', r'\\\1', self._generate_help_text()),
            "    }",
            '}',
            '',
        ])

        lines = [
            '%(whatis_lines)s',
            '',
            "set root %(installdir)s",
        ]

        if self.app.cfg['moduleloadnoconflict']:
            cond_unload = self.conditional_statement(self.is_loaded('%(name)s'), "module unload %(name)s")
            lines.extend([
                '',
                self.conditional_statement(self.is_loaded('%(name)s/%(version)s'), cond_unload, negative=True),
            ])

        elif conflict:
            # conflict on 'name' part of module name (excluding version part at the end)
            # examples:
            # - 'conflict GCC' for 'GCC/4.8.3'
            # - 'conflict Core/GCC' for 'Core/GCC/4.8.2'
            # - 'conflict Compiler/GCC/4.8.2/OpenMPI' for 'Compiler/GCC/4.8.2/OpenMPI/1.6.4'
            lines.extend(['', "conflict %s" % os.path.dirname(self.app.short_mod_name)])

        whatis_lines = [
            "module-whatis {%s}" % re.sub(r'([{}\[\]])', r'\\\1', line)
            for line in self._generate_whatis_lines()
        ]
        txt += '\n'.join([''] + lines + ['']) % {
            'name': self.app.name,
            'version': self.app.version,
            'whatis_lines': '\n'.join(whatis_lines),
            'installdir': self.app.installdir,
        }

        return txt

    def getenv_cmd(self, envvar, default=None):
        """
        Return module-syntax specific code to get value of specific environment variable.
        """
        if default is None:
            cmd = '$::env(%s)' % envvar
        else:
            values = {
                'default': default,
                'envvar': '::env(%s)' % envvar,
            }
            cmd = '[if { [info exists %(envvar)s] } { concat $%(envvar)s } else { concat "%(default)s" } ]' % values
        return cmd

    def load_module(self, mod_name, recursive_unload=None, depends_on=False, unload_modules=None, multi_dep_mods=None):
        """
        Generate load statement for specified module.

        :param mod_name: name of module to generate load statement for
        :param recursive_unload: boolean indicating whether the 'load' statement should be reverted on unload
                                 (if None: enable if recursive_mod_unload build option or depends_on is True)
        :param depends_on: use depends_on statements rather than (guarded) load statements
        :param unload_modules: name(s) of module to unload first
        :param multi_dep_mods: list of module names in multi_deps context, to use for guarding load statement
        """
        body = []
        if unload_modules:
            body.extend([self.unload_module(m).strip() for m in unload_modules])
        load_template = self.LOAD_TEMPLATE
        # Lmod 7.6.1+ supports depends-on which does this most nicely:
        if build_option('mod_depends_on') or depends_on:
            if not self.modules_tool.supports_depends_on:
                raise EasyBuildError("depends-on statements in generated module are not supported by modules tool")
            load_template = self.LOAD_TEMPLATE_DEPENDS_ON

        body.append(load_template)

        depends_on = load_template == self.LOAD_TEMPLATE_DEPENDS_ON

        cond_tmpl = None

        if recursive_unload is None:
            recursive_unload = build_option('recursive_mod_unload') or depends_on

        if recursive_unload:
            # wrapping the 'module load' statement with an 'is-loaded or mode == unload'
            # guard ensures recursive unloading while avoiding load storms;
            # when "module unload" is called on the module in which the
            # dependency "module load" is present, it will get translated
            # to "module unload" (while the condition is left untouched)
            # see also http://lmod.readthedocs.io/en/latest/210_load_storms.html
            cond_tmpl = "[ module-info mode remove ] || %s"

        if depends_on:
            if multi_dep_mods and len(multi_dep_mods) > 1:
                parent_mod_name = os.path.dirname(mod_name)
                guard = self.is_loaded(multi_dep_mods[1:])
                if_body = load_template % {'mod_name': parent_mod_name}
                else_body = '\n'.join(body)
                load_statement = [
                    self.conditional_statement(guard, if_body, else_body=else_body, cond_tmpl=cond_tmpl, cond_or=True),
                ]
            else:
                load_statement = body + ['']
        else:
            if multi_dep_mods is None:
                # guard load statement with check to see whether module being loaded is already loaded
                # (this avoids load storms)
                cond_mod_names = '%(mod_name)s'
            else:
                cond_mod_names = multi_dep_mods

            # conditional load if one or more conditions are specified
            load_guards = self.is_loaded(cond_mod_names)
            body = '\n'.join(body)
            load_statement = [self.conditional_statement(load_guards, body, negative=True, cond_tmpl=cond_tmpl)]

        return '\n'.join([''] + load_statement) % {'mod_name': mod_name}

    def msg_on_load(self, msg):
        """
        Add a message that should be printed when loading the module.
        """
        # escape any (non-escaped) characters with special meaning by prefixing them with a backslash
        msg = re.sub(r'((?<!\\)[%s])' % ''.join(self.CHARS_TO_ESCAPE), r'\\\1', msg)
        print_cmd = "puts stderr %s" % quote_str(msg, tcl=True)
        return '\n'.join(['', self.conditional_statement("module-info mode load", print_cmd, indent=False)])

    def update_paths(self, key, paths, prepend=True, allow_abs=False, expand_relpaths=True):
        """
        Generate prepend-path or append-path statements for the given list of paths.

        :param key: environment variable to prepend/append paths to
        :param paths: list of paths to prepend
        :param prepend: whether to prepend (True) or append (False) paths
        :param allow_abs: allow providing of absolute paths
        :param expand_relpaths: expand relative paths into absolute paths (by prefixing install dir)
        """
        if prepend:
            update_type = 'prepend'
        else:
            update_type = 'append'

        if not self.define_env_var(key):
            self.log.info("Not including statement to %s environment variable $%s, as specified", update_type, key)
            return ''

        if isinstance(paths, string_type):
            self.log.debug("Wrapping %s into a list before using it to %s path %s", paths, update_type, key)
            paths = [paths]

        abspaths = []
        for path in paths:
            if os.path.isabs(path) and not allow_abs:
                raise EasyBuildError("Absolute path %s passed to update_paths which only expects relative paths.",
                                     path)
            elif not os.path.isabs(path):
                # prepend/append $root (= installdir) for (non-empty) relative paths
                if path:
                    if expand_relpaths:
                        abspaths.append(os.path.join('$root', path))
                    else:
                        abspaths.append(path)
                else:
                    abspaths.append('$root')
            else:
                abspaths.append(path)

        statements = ['%s-path\t%s\t\t%s\n' % (update_type, key, p) for p in abspaths]
        return ''.join(statements)

    def set_alias(self, key, value):
        """
        Generate set-alias statement in modulefile for the given key/value pair.
        """
        # quotes are needed, to ensure smooth working of EBDEVEL* modulefiles
        return 'set-alias\t%s\t\t%s\n' % (key, quote_str(value, tcl=True))

    def set_as_default(self, module_dir_path, module_version, mod_symlink_paths=None):
        """
        Create a .version file inside the package module folder in order to set the default version for TMod

        :param module_dir_path: module directory path, e.g. $HOME/easybuild/modules/all/Bison
        :param module_version: module version, e.g. 3.0.4
        :param mod_symlink_paths: list of paths in which symlinks to module files must be created
        """
        txt = self.MODULE_SHEBANG + '\n'
        txt += 'set ModulesVersion %s\n' % module_version

        # write the file no matter what
        dot_version_path = os.path.join(module_dir_path, '.version')
        write_file(dot_version_path, txt)

        # create symlink to .version file in class module folders
        if mod_symlink_paths is None:
            mod_symlink_paths = []

        module_dir_name = os.path.basename(module_dir_path)
        for mod_symlink_path in mod_symlink_paths:
            mod_symlink_dir = os.path.join(install_path('mod'), mod_symlink_path, module_dir_name)
            dot_version_link_path = os.path.join(mod_symlink_dir, '.version')
            if os.path.islink(dot_version_link_path):
                link_target = resolve_path(dot_version_link_path)
                remove_file(dot_version_link_path)
                self.log.info("Removed default version marking from %s.", link_target)
            elif os.path.exists(dot_version_link_path):
                raise EasyBuildError('Found an unexpected file named .version in dir %s', mod_symlink_dir)
            symlink(dot_version_path, dot_version_link_path, use_abspath_source=True)

    def set_environment(self, key, value, relpath=False):
        """
        Generate a quoted setenv statement for the given key/value pair.

        :param key: name of environment variable to define
        :param value: value to define environment variable with
        :param relpath: value is path relative to installation prefix
        """
        if not self.define_env_var(key):
            self.log.info("Not including statement to define environment variable $%s, as specified", key)
            return ''

        # quotes are needed, to ensure smooth working of EBDEVEL* modulefiles
        if relpath:
            if value:
                val = quote_str(os.path.join('$root', value), tcl=True)
            else:
                val = '"$root"'
        else:
            val = quote_str(value, tcl=True)
        return 'setenv\t%s\t\t%s\n' % (key, val)

    def swap_module(self, mod_name_out, mod_name_in, guarded=True):
        """
        Generate swap statement for specified module names.

        :param mod_name_out: name of module to unload (swap out)
        :param mod_name_in: name of module to load (swap in)
        :param guarded: guard 'swap' statement, fall back to 'load' if module being swapped out is not loaded
        """
        # In Modules 4.2.3+ a 2-argument swap 'module swap foo foo/X.Y.Z' will fail as the unloaded 'foo'
        # means all 'foo' modules conflict and 'foo/X.Y.Z' will not load.  A 1-argument swap like
        # 'module swap foo/X.Y.Z' will unload any currently loaded 'foo' without it becoming conflicting
        # and successfully load the new module.
        # See: https://modules.readthedocs.io/en/latest/NEWS.html#modules-4-2-3-2019-03-23
        body = "module swap %s" % (mod_name_in)
        if guarded:
            alt_body = self.LOAD_TEMPLATE % {'mod_name': mod_name_in}
            swap_statement = [self.conditional_statement(self.is_loaded(mod_name_out), body, else_body=alt_body)]
        else:
            swap_statement = [body, '']

        return '\n'.join([''] + swap_statement)

    def unload_module(self, mod_name):
        """
        Generate unload statement for specified module.

        :param mod_name: name of module to generate unload statement for
        """
        return '\n'.join(['', "module unload %s" % mod_name])

    def det_user_modpath(self, user_modpath):
        """
        Determine user-specific modules subdirectory, to be used in 'use' statements
        (cfr. implementation of use() method).
        """
        if user_modpath:
            user_modpath = ' '.join(self._det_user_modpath_common(user_modpath))

        return user_modpath

    def use(self, paths, prefix=None, guarded=False, user_modpath=None):
        """
        Generate module use statements for given list of module paths.
        :param paths: list of module path extensions to generate use statements for; paths will be quoted
        :param prefix: optional path prefix; not quoted, i.e., can be a statement
        :param guarded: use statements will be guarded to only apply if path exists
        :param user_modpath: user-specific modules subdirectory to include in use statements
        """
        user_modpath = self.det_user_modpath(user_modpath)
        use_statements = []
        for path in paths:
            quoted_path = quote_str(path, tcl=True)
            if user_modpath:
                quoted_path = '[ file join %s %s ]' % (user_modpath, quoted_path)
            if prefix:
                full_path = '[ file join %s %s ]' % (prefix, quoted_path)
            else:
                full_path = quoted_path
            if guarded:
                cond_statement = self.conditional_statement('file isdirectory %s' % full_path,
                                                            'module use %s' % full_path)
                use_statements.append(cond_statement)
            else:
                use_statements.append("module use %s\n" % full_path)
        return ''.join(use_statements)


class ModuleGeneratorLua(ModuleGenerator):
    """
    Class for generating Lua module files.
    """
    SYNTAX = 'Lua'
    MODULE_FILE_EXTENSION = '.lua'
    MODULE_SHEBANG = ''  # no 'shebang' in Lua module files
    CHARS_TO_ESCAPE = []

    INSTALLDIR_REGEX = r'^local root\s+=\s+"(?P<installdir>.*)"'
    LOAD_REGEX = r'^\s*(?:load|depends_on)\("(\S+)"'
    LOAD_TEMPLATE = 'load("%(mod_name)s")'
    LOAD_TEMPLATE_DEPENDS_ON = 'depends_on("%(mod_name)s")'
    IS_LOADED_TEMPLATE = 'isloaded("%s")'

    PATH_JOIN_TEMPLATE = 'pathJoin(root, "%s")'
    UPDATE_PATH_TEMPLATE = '%s_path("%s", %s)'

    START_STR = '[==['
    END_STR = ']==]'

    def __init__(self, *args, **kwargs):
        """ModuleGeneratorLua constructor."""
        super(ModuleGeneratorLua, self).__init__(*args, **kwargs)

        if self.modules_tool:
            if self.modules_tool.version and LooseVersion(self.modules_tool.version) >= LooseVersion('7.7.38'):
                self.DOT_MODULERC = '.modulerc.lua'

    def check_version(self, minimal_version_maj, minimal_version_min, minimal_version_patch='0'):
        """
        Check the minimal version of the moduletool in the module file
        :param minimal_version_maj: the major version to check
        :param minimal_version_min: the minor version to check
        :param minimal_version_patch: the patch version to check
        """
        lmod_version_check_expr = 'convertToCanonical(LmodVersion()) >= convertToCanonical("%(maj)s.%(min)s.%(patch)s")'
        return lmod_version_check_expr % {
            'maj': minimal_version_maj,
            'min': minimal_version_min,
            'patch': minimal_version_patch,
        }

    def check_group(self, group, error_msg=None):
        """
        Generate a check of the software group and the current user, and refuse to load the module if the user don't
        belong to the group

        :param group: string with the group name
        :param error_msg: error message to print for users outside that group
        """
        lmod_version = self.modules_tool.version
        min_lmod_version = '6.0.8'

        if LooseVersion(lmod_version) >= LooseVersion(min_lmod_version):
            if error_msg is None:
                error_msg = "You are not part of '%s' group of users that have access to this software; " % group
                error_msg += "Please consult with user support how to become a member of this group"

            error_msg = 'LmodError("' + error_msg + '")'
            res = self.conditional_statement('userInGroup("%s")' % group, error_msg, negative=True)
        else:
            warn_msg = "Can't generate robust check in Lua modules for users belonging to group %s. "
            warn_msg += "Lmod version not recent enough (%s), should be >= %s"
            self.log.warning(warn_msg, group, lmod_version, min_lmod_version)
            res = ''

        return res

    def check_str(self, txt):
        """Check whether provided string has any unwanted substrings in it."""
        if self.START_STR in txt or self.END_STR in txt:
            raise EasyBuildError("Found unwanted '%s' or '%s' in: %s", self.START_STR, self.END_STR, txt)
        else:
            return txt

    def comment(self, msg):
        """Return string containing given message as a comment."""
        return "-- %s\n" % msg

    def conditional_statement(self, conditions, body, negative=False, else_body=None, indent=True,
                              cond_or=False, cond_tmpl=None):
        """
        Return formatted conditional statement, with given condition and body.

        :param conditions: (list of) string(s) containing the statement(s) for the if condition (in correct syntax)
        :param body: (multiline) string with if body (in correct syntax, without indentation)
        :param negative: boolean indicating whether the (individual) condition(s) should be negated
        :param else_body: optional body for 'else' part
        :param indent: indent if/else body
        :param cond_or: combine multiple conditions using 'or' (default is to combine with 'and')
        :param cond_tmpl: template for condition expression (default: '%s')
        """
        if isinstance(conditions, string_type):
            conditions = [conditions]

        if cond_or:
            join_op = ' or '
        else:
            join_op = ' and '

        if negative:
            condition = join_op.join('not ( %s )' % c for c in conditions)
        else:
            condition = join_op.join(conditions)

        if cond_tmpl:
            condition = cond_tmpl % condition

        lines = ["if %s then" % condition]

        for line in body.split('\n'):
            if indent:
                line = self.INDENTATION + line
            lines.append(line)

        if else_body is None:
            lines.extend(['end', ''])
        else:
            lines.append('else')
            for line in else_body.split('\n'):
                if indent:
                    line = self.INDENTATION + line
                lines.append(line)
            lines.extend(['end', ''])

        return '\n'.join(lines)

    def get_description(self, conflict=True):
        """
        Generate a description.
        """
        txt = '\n'.join([
            'help(%s%s' % (self.START_STR, self.check_str(self._generate_help_text())),
            '%s)' % self.END_STR,
            '',
        ])

        lines = [
            "%(whatis_lines)s",
            '',
            'local root = "%(installdir)s"',
        ]

        if self.app.cfg['moduleloadnoconflict']:
            self.log.info("Nothing to do to ensure no conflicts can occur on load when using Lua modules files/Lmod")

        elif conflict:
            # conflict on 'name' part of module name (excluding version part at the end)
            lines.extend(['', 'conflict("%s")' % os.path.dirname(self.app.short_mod_name)])

        whatis_lines = []
        for line in self._generate_whatis_lines():
            whatis_lines.append("whatis(%s%s%s)" % (self.START_STR, self.check_str(line), self.END_STR))

        if build_option('module_extensions'):
            extensions_list = self._generate_extensions_list()

            if extensions_list:
                extensions_stmt = 'extensions("%s")' % ','.join(['%s' % x for x in extensions_list])
                # put this behind a Lmod version check as 'extensions' is only (well) supported since Lmod 8.2.8,
                # see https://lmod.readthedocs.io/en/latest/330_extensions.html#module-extensions and
                # https://github.com/TACC/Lmod/issues/428
                lines.extend(['', self.conditional_statement(self.check_version("8", "2", "8"), extensions_stmt)])

        txt += '\n'.join([''] + lines + ['']) % {
            'name': self.app.name,
            'version': self.app.version,
            'whatis_lines': '\n'.join(whatis_lines),
            'installdir': self.app.installdir,
            'homepage': self.app.cfg['homepage'],
        }

        return txt

    def getenv_cmd(self, envvar, default=None):
        """
        Return module-syntax specific code to get value of specific environment variable.
        """
        if default is None:
            cmd = 'os.getenv("%s")' % envvar
        else:
            cmd = 'os.getenv("%s") or "%s"' % (envvar, default)
        return cmd

    def load_module(self, mod_name, recursive_unload=None, depends_on=False, unload_modules=None, multi_dep_mods=None):
        """
        Generate load statement for specified module.

        :param mod_name: name of module to generate load statement for
        :param recursive_unload: boolean indicating whether the 'load' statement should be reverted on unload
                                 (if None: enable if recursive_mod_unload build option or depends_on is True)
        :param depends_on: use depends_on statements rather than (guarded) load statements
        :param unload_modules: name(s) of module to unload first
        :param multi_dep_mods: list of module names in multi_deps context, to use for guarding load statement
        """
        body = []
        if unload_modules:
            body.extend([self.unload_module(m).strip() for m in unload_modules])

        load_template = self.LOAD_TEMPLATE
        # Lmod 7.6+ supports depends_on which does this most nicely:
        if build_option('mod_depends_on') or depends_on:
            if not self.modules_tool.supports_depends_on:
                raise EasyBuildError("depends_on statements in generated module are not supported by modules tool")
            load_template = self.LOAD_TEMPLATE_DEPENDS_ON

        body.append(load_template)

        depends_on = load_template == self.LOAD_TEMPLATE_DEPENDS_ON

        cond_tmpl = None

        if recursive_unload is None:
            recursive_unload = build_option('recursive_mod_unload') or depends_on

        if recursive_unload:
            # wrapping the 'module load' statement with an 'is-loaded or mode == unload'
            # guard ensures recursive unloading while avoiding load storms;
            # when "module unload" is called on the module in which the
            # dependency "module load" is present, it will get translated
            # to "module unload" (while the condition is left untouched)
            # see also http://lmod.readthedocs.io/en/latest/210_load_storms.html
            cond_tmpl = 'mode() == "unload" or %s'

        if depends_on:
            if multi_dep_mods and len(multi_dep_mods) > 1:
                parent_mod_name = os.path.dirname(mod_name)
                guard = self.is_loaded(multi_dep_mods[1:])
                if_body = load_template % {'mod_name': parent_mod_name}
                else_body = '\n'.join(body)
                load_statement = [
                    self.conditional_statement(guard, if_body, else_body=else_body, cond_tmpl=cond_tmpl, cond_or=True),
                ]
            else:
                load_statement = body + ['']
        else:
            if multi_dep_mods is None:
                # guard load statement with check to see whether module being loaded is already loaded
                # (this avoids load storms)
                cond_mod_names = '%(mod_name)s'
            else:
                cond_mod_names = multi_dep_mods

            # conditional load if one or more conditions are specified
            load_guards = self.is_loaded(cond_mod_names)
            body = '\n'.join(body)
            load_statement = [self.conditional_statement(load_guards, body, negative=True, cond_tmpl=cond_tmpl)]

        return '\n'.join([''] + load_statement) % {'mod_name': mod_name}

    def msg_on_load(self, msg):
        """
        Add a message that should be printed when loading the module.
        """
        # take into account possible newlines in messages by using [==...==] (requires Lmod 5.8)
        stmt = 'io.stderr:write(%s%s%s)' % (self.START_STR, self.check_str(msg), self.END_STR)
        return '\n' + self.conditional_statement('mode() == "load"', stmt, indent=False)

    def modulerc(self, module_version=None, filepath=None, modulerc_txt=None):
        """
        Generate contents of .modulerc(.lua) file, in Lua syntax (but only if Lmod is recent enough, i.e. >= 7.7.38)

        :param module_version: specs for module-version statement (dict with 'modname', 'sym_version' & 'version' keys)
        :param filepath: location where .modulerc file should be written to
        :param modulerc_txt: contents of .modulerc to use
        :return: contents of .modulerc file
        """
        if modulerc_txt is None:
            lmod_ver = self.modules_tool.version
            min_ver = '7.7.38'

            if LooseVersion(lmod_ver) >= LooseVersion(min_ver):
                self.log.info("Found Lmod v%s >= v%s, so will generate .modulerc.lua in Lua syntax", lmod_ver, min_ver)

                modulerc = []

                if self._modulerc_check_module_version(module_version):
                    module_version_statement = 'module_version("%(modname)s", "%(sym_version)s")'
                    modulerc.append(module_version_statement % module_version)

                modulerc_txt = '\n'.join(modulerc)

            else:
                self.log.info("Lmod v%s < v%s, need to stick to Tcl syntax for .modulerc", lmod_ver, min_ver)

        return super(ModuleGeneratorLua, self).modulerc(module_version=module_version, filepath=filepath,
                                                        modulerc_txt=modulerc_txt)

    def update_paths(self, key, paths, prepend=True, allow_abs=False, expand_relpaths=True):
        """
        Generate prepend_path or append_path statements for the given list of paths

        :param key: environment variable to prepend/append paths to
        :param paths: list of paths to prepend/append
        :param prepend: whether to prepend (True) or append (False) paths
        :param allow_abs: allow providing of absolute paths
        :param expand_relpaths: expand relative paths into absolute paths (by prefixing install dir)
        """
        if prepend:
            update_type = 'prepend'
        else:
            update_type = 'append'

        if not self.define_env_var(key):
            self.log.info("Not including statement to %s environment variable $%s, as specified", update_type, key)
            return ''

        if isinstance(paths, string_type):
            self.log.debug("Wrapping %s into a list before using it to %s path %s", update_type, paths, key)
            paths = [paths]

        abspaths = []
        for path in paths:
            if os.path.isabs(path):
                if allow_abs:
                    abspaths.append(quote_str(path))
                else:
                    raise EasyBuildError("Absolute path %s passed to update_paths which only expects relative paths.",
                                         path)
            else:
                # use pathJoin for (non-empty) relative paths
                if path:
                    if expand_relpaths:
                        abspaths.append(self.PATH_JOIN_TEMPLATE % path)
                    else:
                        abspaths.append(quote_str(path))
                else:
                    abspaths.append('root')

        statements = [self.UPDATE_PATH_TEMPLATE % (update_type, key, p) for p in abspaths]
        statements.append('')
        return '\n'.join(statements)

    def set_alias(self, key, value):
        """
        Generate set-alias statement in modulefile for the given key/value pair.
        """
        # quotes are needed, to ensure smooth working of EBDEVEL* modulefiles
        return 'set_alias("%s", %s)\n' % (key, quote_str(value))

    def set_as_default(self, module_dir_path, module_version, mod_symlink_paths=None):
        """
        Create a symlink named 'default' inside the package's module folder in order to set the default module version

        :param module_dir_path: module directory path, e.g. $HOME/easybuild/modules/all/Bison
        :param module_version: module version, e.g. 3.0.4
        :param mod_symlink_paths: list of paths in which symlinks to module files must be created
        """
        def create_default_symlink(path):
            """Helper function to create 'default' symlink in specified directory."""
            default_filepath = os.path.join(path, 'default')

            if os.path.islink(default_filepath):
                link_target = resolve_path(default_filepath)
                remove_file(default_filepath)
                self.log.info("Removed default version marking from %s.", link_target)
            elif os.path.exists(default_filepath):
                raise EasyBuildError('Found an unexpected file named default in dir %s', module_dir_path)

            symlink(module_version + self.MODULE_FILE_EXTENSION, default_filepath, use_abspath_source=False)
            self.log.info("Module default version file written to point to %s", default_filepath)

        create_default_symlink(module_dir_path)

        # also create symlinks in class module folders
        if mod_symlink_paths is None:
            mod_symlink_paths = []

        for mod_symlink_path in mod_symlink_paths:
            mod_dir_name = os.path.basename(module_dir_path)
            mod_symlink_dir = os.path.join(install_path('mod'), mod_symlink_path, mod_dir_name)
            create_default_symlink(mod_symlink_dir)

    def set_environment(self, key, value, relpath=False):
        """
        Generate a quoted setenv statement for the given key/value pair.

        :param key: name of environment variable to define
        :param value: value to define environment variable with
        :param relpath: value is path relative to installation prefix
        """
        if not self.define_env_var(key):
            self.log.info("Not including statement to define environment variable $%s, as specified", key)
            return ''

        if relpath:
            if value:
                val = self.PATH_JOIN_TEMPLATE % value
            else:
                val = 'root'
        else:
            val = quote_str(value)
        return 'setenv("%s", %s)\n' % (key, val)

    def swap_module(self, mod_name_out, mod_name_in, guarded=True):
        """
        Generate swap statement for specified module names.

        :param mod_name_out: name of module to unload (swap out)
        :param mod_name_in: name of module to load (swap in)
        :param guarded: guard 'swap' statement, fall back to 'load' if module being swapped out is not loaded
        """
        body = '\n'.join([
            'unload("%s")' % mod_name_out,
            'load("%s")' % mod_name_in,
        ])
        if guarded:
            alt_body = self.LOAD_TEMPLATE % {'mod_name': mod_name_in}
            swap_statement = [self.conditional_statement(self.is_loaded(mod_name_out), body, else_body=alt_body)]
        else:
            swap_statement = [body, '']

        return '\n'.join([''] + swap_statement)

    def unload_module(self, mod_name):
        """
        Generate unload statement for specified module.

        :param mod_name: name of module to generate unload statement for
        """
        return '\n'.join(['', 'unload("%s")' % mod_name])

    def det_user_modpath(self, user_modpath):
        """
        Determine user-specific modules subdirectory, to be used in 'use' statements
        (cfr. implementations of use() method).
        """
        if user_modpath:
            user_modpath = ', '.join(self._det_user_modpath_common(user_modpath))

        return user_modpath

    def use(self, paths, prefix=None, guarded=False, user_modpath=None):
        """
        Generate module use statements for given list of module paths.
        :param paths: list of module path extensions to generate use statements for; paths will be quoted
        :param prefix: optional path prefix; not quoted, i.e., can be a statement
        :param guarded: use statements will be guarded to only apply if path exists
        :param user_modpath: user-specific modules subdirectory to include in use statements
        """
        user_modpath = self.det_user_modpath(user_modpath)
        use_statements = []
        for path in paths:
            quoted_path = quote_str(path)
            if user_modpath:
                quoted_path = 'pathJoin(%s, %s)' % (user_modpath, quoted_path)
            if prefix:
                full_path = 'pathJoin(%s, %s)' % (prefix, quoted_path)
            else:
                full_path = quoted_path

            prepend_modulepath = self.UPDATE_PATH_TEMPLATE % ('prepend', 'MODULEPATH', full_path)
            if guarded:
                cond_statement = self.conditional_statement('isDir(%s)' % full_path, prepend_modulepath)
                use_statements.append(cond_statement)
            else:
                use_statements.append(prepend_modulepath + '\n')
        return ''.join(use_statements)
